#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Provides access functions for the app identity service.

To learn more about the App Identity API, review the Overview document:
https://cloud.google.com/appengine/docs/python/appidentity/
"""










import os
import time

from google.appengine.api import apiproxy_stub_map
from google.appengine.api import full_app_id
from google.appengine.api.app_identity import _metadata_server
from google.appengine.api.app_identity import app_identity_service_pb2
from google.appengine.runtime import apiproxy_errors
import six

__all__ = ['BackendDeadlineExceeded',
           'BlobSizeTooLarge',
           'InternalError',
           'InvalidScope',
           'NotAllowed',
           'OperationNotImplemented',
           'Error',
           'create_rpc',
           'make_sign_blob_call',
           'make_get_public_certificates_call',
           'make_get_service_account_name_call',
           'sign_blob',
           'get_public_certificates',
           'PublicCertificate',
           'get_service_account_name',
           'get_application_id',
           'get_default_version_hostname',
           'get_access_token',
           'get_default_gcs_bucket_name',
           'make_get_default_gcs_bucket_name_call',
          ]


_APP_IDENTITY_SERVICE_NAME = 'app_identity_service'
_SIGN_FOR_APP_METHOD_NAME = 'SignForApp'
_GET_CERTS_METHOD_NAME = 'GetPublicCertificatesForApp'
_GET_SERVICE_ACCOUNT_NAME_METHOD_NAME = 'GetServiceAccountName'
_GET_DEFAULT_GCS_BUCKET_NAME_METHOD_NAME = 'GetDefaultGcsBucketName'



_TOKEN_EXPIRY_SAFETY_MARGIN = 300
_MAX_TOKEN_CACHE_SIZE = 100


_MAX_RANDOM_EXPIRY_DELTA = 60





_random_cache_expiry_delta = (
    hash(time.time()) % (_MAX_RANDOM_EXPIRY_DELTA * 1000) / 1000.0)


class Error(Exception):
  """Base error type."""



class BackendDeadlineExceeded(Error):
  """The communication to the backend service timed out."""


class BlobSizeTooLarge(Error):
  """The size of the blob to sign is larger than the allowed limit."""


class InternalError(Error):
  """An unspecified internal failure occurred."""


class InvalidScope(Error):
  """The scope is invalid."""


class NotAllowed(Error):
  """The operation is not allowed."""


class OperationNotImplemented(Error):
  """The operation is not implemented for the service account."""


def _to_app_identity_error(error):
  """Translates an application error to an external error, if possible.

  Args:
    error: An `ApplicationError` to translate.

  Returns:
    error: The App Identity API-specific error message.
  """
  error_map = {
      app_identity_service_pb2.AppIdentityServiceError.NOT_A_VALID_APP:
      InternalError,
      app_identity_service_pb2.AppIdentityServiceError.DEADLINE_EXCEEDED:
      BackendDeadlineExceeded,
      app_identity_service_pb2.AppIdentityServiceError.BLOB_TOO_LARGE:
      BlobSizeTooLarge,
      app_identity_service_pb2.AppIdentityServiceError.UNKNOWN_ERROR:
      InternalError,
      app_identity_service_pb2.AppIdentityServiceError.UNKNOWN_SCOPE:
      InvalidScope,
      app_identity_service_pb2.AppIdentityServiceError.NOT_ALLOWED:
      NotAllowed,
      app_identity_service_pb2.AppIdentityServiceError.NOT_IMPLEMENTED:
      OperationNotImplemented,
      }
  if error.application_error in error_map:
    return error_map[error.application_error](error.error_detail)
  else:
    return InternalError('%s: %s' %
                         (error.application_error, error.error_detail))


class PublicCertificate(object):
  """Class that specifies information about a public certificate.

  Attributes:
    key_name: Name of the certificate.
    x509_certificate_pem: X.509
        certificates (https://www.ietf.org/rfc/rfc2459.txt) in PEM format.


  """

  def __init__(self, key_name, x509_certificate_pem):
    self.key_name = key_name
    self.x509_certificate_pem = x509_certificate_pem


def create_rpc(deadline=None, callback=None):
  """Creates an RPC object for use with the App Identity API.

  Args:
    deadline: Optional deadline in seconds for the operation; the default value
        is a system-specific deadline, typically 5 seconds.
    callback: Optional callable to invoke on completion.

  Returns:
    An `apiproxy_stub_map.UserRPC` object specialized for this service.
  """
  return apiproxy_stub_map.UserRPC(_APP_IDENTITY_SERVICE_NAME,
                                   deadline, callback)


def make_sign_blob_call(rpc, bytes_to_sign):
  """Executes the RPC call to sign a blob.

  Args:
    rpc: A `UserRPC` instance.
    bytes_to_sign: Blob that must be signed.

  Returns:
    A tuple that contains the signing key name and the signature.

  Raises:
    TypeError: If `bytes_to_sign` is not a string or bytes type.
  """
  if not isinstance(bytes_to_sign, (six.text_type, bytes)):
    raise TypeError('bytes_to_sign must be str or bytes: %s' % bytes_to_sign)
  request = app_identity_service_pb2.SignForAppRequest()
  request.bytes_to_sign = six.ensure_binary(bytes_to_sign)
  response = app_identity_service_pb2.SignForAppResponse()

  def signing_for_app_result(rpc):
    """Checks success, handles exceptions, and returns the converted RPC result.

    This method waits for the RPC if it has not yet finished and calls the
    post-call hooks on the first invocation.

    Args:
      rpc: A `UserRPC` object.

    Returns:
      A tuple that contains signing key name and signature.
    """
    assert rpc.service == _APP_IDENTITY_SERVICE_NAME, repr(rpc.service)
    assert rpc.method == _SIGN_FOR_APP_METHOD_NAME, repr(rpc.method)
    try:
      rpc.check_success()
    except apiproxy_errors.ApplicationError as err:
      raise _to_app_identity_error(err)

    return (response.key_name, response.signature_bytes)


  rpc.make_call(_SIGN_FOR_APP_METHOD_NAME, request,
                response, signing_for_app_result)


def make_get_public_certificates_call(rpc):
  """Executes the RPC call to get a list of public certificates.

  Args:
    rpc: A `UserRPC` instance.

  Returns:
    A list of `PublicCertificate` objects.
  """
  request = app_identity_service_pb2.GetPublicCertificateForAppRequest()
  response = app_identity_service_pb2.GetPublicCertificateForAppResponse()

  def get_certs_result(rpc):
    """Checks success, handles exceptions, and returns the converted RPC result.

    This method waits for the RPC if it has not yet finished and calls the
    post-call hooks on the first invocation.

    Args:
      rpc: A `UserRPC` object.

    Returns:
      A list of `PublicCertificate` objects.
    """
    assert rpc.service == _APP_IDENTITY_SERVICE_NAME, repr(rpc.service)
    assert rpc.method == _GET_CERTS_METHOD_NAME, repr(rpc.method)
    try:
      rpc.check_success()
    except apiproxy_errors.ApplicationError as err:
      raise _to_app_identity_error(err)
    result = []
    for cert in response.public_certificate_list:
      result.append(PublicCertificate(
          cert.key_name, cert.x509_certificate_pem))
    return result


  rpc.make_call(_GET_CERTS_METHOD_NAME, request, response, get_certs_result)


def make_get_service_account_name_call(rpc):
  """Gets the service account name of the app.

  Args:
    rpc: A `UserRPC` object.

  Returns:
    Service account name of the app.
  """
  request = app_identity_service_pb2.GetServiceAccountNameRequest()
  response = app_identity_service_pb2.GetServiceAccountNameResponse()

  def get_service_account_name_result(rpc):
    """Checks success, handles exceptions, and returns the converted RPC result.

    This method waits for the RPC if it has not yet finished and calls the
    post-call hooks on the first invocation.

    Args:
      rpc: A `UserRPC` object.

    Returns:
      A string of the service account name of the app.
    """
    assert rpc.service == _APP_IDENTITY_SERVICE_NAME, repr(rpc.service)
    assert rpc.method == _GET_SERVICE_ACCOUNT_NAME_METHOD_NAME, repr(rpc.method)
    try:
      rpc.check_success()
    except apiproxy_errors.ApplicationError as err:
      raise _to_app_identity_error(err)

    return response.service_account_name


  rpc.make_call(_GET_SERVICE_ACCOUNT_NAME_METHOD_NAME, request,
                response, get_service_account_name_result)


def make_get_default_gcs_bucket_name_call(rpc):
  """Gets the default Google Cloud Storage bucket name for the app.

  Args:
    rpc: A `UserRPC` object.

  Returns:
    The default Google Cloud Storage bucket name for the app.
  """
  request = app_identity_service_pb2.GetDefaultGcsBucketNameRequest()
  response = app_identity_service_pb2.GetDefaultGcsBucketNameResponse()

  if rpc.deadline is not None:
    request.deadline = rpc.deadline

  def get_default_gcs_bucket_name_result(rpc):
    """Checks success, handles exceptions, and returns the converted RPC result.

    This method waits for the RPC if it has not yet finished and calls the
    post-call hooks on the first invocation.

    Args:
      rpc: A `UserRPC` object.

    Returns:
      A string of the name of the app's default Google Cloud Storage bucket.
    """
    assert rpc.service == _APP_IDENTITY_SERVICE_NAME, repr(rpc.service)
    assert rpc.method == _GET_DEFAULT_GCS_BUCKET_NAME_METHOD_NAME, (
        repr(rpc.method))
    try:
      rpc.check_success()
    except apiproxy_errors.ApplicationError as err:
      raise _to_app_identity_error(err)

    return response.default_gcs_bucket_name or None


  rpc.make_call(_GET_DEFAULT_GCS_BUCKET_NAME_METHOD_NAME, request,
                response, get_default_gcs_bucket_name_result)


def sign_blob(bytes_to_sign, deadline=None):
  """Signs a blob.

  Args:
    bytes_to_sign: The blob that must be signed.
    deadline: Optional deadline in seconds for the operation; the default value
      is a system-specific deadline, typically 5 seconds.

  Returns:
    A tuple containing the signing key name and signature.
  """
  rpc = create_rpc(deadline)
  make_sign_blob_call(rpc, bytes_to_sign)
  rpc.wait()
  return rpc.get_result()


def get_public_certificates(deadline=None):
  """Gets public certificates.

  Args:
    deadline: Optional deadline in seconds for the operation; the default value
        is a system-specific deadline, typically 5 seconds.

  Returns:
    A list of `PublicCertificate` objects.
  """
  rpc = create_rpc(deadline)
  make_get_public_certificates_call(rpc)
  rpc.wait()
  return rpc.get_result()


def get_service_account_name(deadline=None):
  """Gets the service account name of the app.

  Args:
    deadline: Optional deadline in seconds for the operation; the default value
        is a system-specific deadline, typically 5 seconds.

  Returns:
    The service account name of the app.
  """
  rpc = create_rpc(deadline)
  make_get_service_account_name_call(rpc)
  rpc.wait()
  return rpc.get_result()


def get_default_gcs_bucket_name(deadline=None):
  """Gets the default Google Cloud Storage bucket name for the app.

  Args:
    deadline: Optional deadline in seconds for the operation; the default value
        is a system-specific deadline, typically 5 seconds.

  Returns:
    Default bucket name for the app.
  """
  rpc = create_rpc(deadline)
  make_get_default_gcs_bucket_name_call(rpc)
  rpc.wait()
  return rpc.get_result()


def get_application_id():
  """Gets the application ID of an app.

  Not to be confused with `google.appengine.api.full_app_id.get()`, which gets
  the "raw" app ID.

  Returns:
    The application ID of the app.
  """
  return full_app_id.project_id()


def get_default_version_hostname():
  """Gets the standard host name of the default version of the app.

  For example, if your `application_id` is `my-app`, then the result might be
  `my-app.appspot.com`.

  Returns:
    The standard host name of the default version of the application.
  """





  return os.getenv('DEFAULT_VERSION_HOSTNAME')




def get_access_token(scopes, service_account_id=None):
  """The OAuth 2.0 access token to act on behalf of the application.

  Each application has an associated Google account. This function returns an
  OAuth 2.0 access token that corresponds to the running app. Access tokens are
  safe to cache and reuse until their expiry time as returned.

  Args:
    scopes: The requested API scope string, or a list of strings.

  Returns:
    A `Pair`, `Access` token string and the expiration time in seconds since
    the epoch.

  Raises:
    InvalidScope: If the scopes are unspecified or invalid.
  """
  if not scopes:
    raise InvalidScope('No scopes specified.')
  if isinstance(scopes, six.string_types):
    scopes = [scopes]
  return _metadata_server.get_service_account_token(
      scopes=scopes, service_account=service_account_id)


_ParseFullAppId = full_app_id.parse
_PARTITION_SEPARATOR = '~'
_DOMAIN_SEPARATOR = ':'



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#



"""Higher-level, semantic data types for the datastore.

These types are expected to be set as attributes of `Entities`.

Most of these types are based on XML elements from Atom and GData elements
from the `atom` and `gd` namespaces. For more information, see:

- http://www.atomenabled.org/developers/syndication/
- https://developers.google.com/gdata/docs/1.0/elements

The namespace schemas are:

- http://www.w3.org/2005/Atom
- http://schemas.google.com/g/2005
"""








import array
import base64
import binascii
import calendar
import datetime
import os
import re
import struct
import time
from xml.sax import saxutils

from google.appengine.api import cmp_compat
from google.appengine.api import datastore_errors
from google.appengine.api import full_app_id
from google.appengine.api import namespace_manager
from google.appengine.api import users
from google.appengine.datastore import datastore_pbs
from google.appengine.datastore import entity_v4_pb2
from google.appengine.datastore import sortable_pb_encoder
import six
from six.moves import range
from six.moves import urllib
from six.moves import zip

from google.appengine.datastore import entity_bytes_pb2 as entity_pb2









if datastore_pbs._CLOUD_DATASTORE_ENABLED:
  from google.appengine.datastore.datastore_pbs import googledatastore







_MAX_STRING_LENGTH = 1500











_MAX_LINK_PROPERTY_LENGTH = 2083




_MAX_RAW_PROPERTY_BYTES = 1048487







RESERVED_PROPERTY_NAME = re.compile('^__.*__$')







KEY_SPECIAL_PROPERTY = '__key__'
_KEY_SPECIAL_PROPERTY = KEY_SPECIAL_PROPERTY
_UNAPPLIED_LOG_TIMESTAMP_SPECIAL_PROPERTY = '__unapplied_log_timestamp_us__'
SCATTER_SPECIAL_PROPERTY = '__scatter__'
_SPECIAL_PROPERTIES = frozenset(
    [KEY_SPECIAL_PROPERTY,
     _UNAPPLIED_LOG_TIMESTAMP_SPECIAL_PROPERTY,
     SCATTER_SPECIAL_PROPERTY])







_NAMESPACE_SEPARATOR = '!'





_EMPTY_NAMESPACE_ID = 1


_EPOCH = datetime.datetime.utcfromtimestamp(0)


if six.PY2:
  _PREFERRED_NUM_TYPE = long
else:
  _PREFERRED_NUM_TYPE = int





class UtcTzinfo(datetime.tzinfo):
  def utcoffset(self, dt): return datetime.timedelta(0)
  def dst(self, dt): return datetime.timedelta(0)
  def tzname(self, dt): return 'UTC'
  def __repr__(self): return 'datastore_types.UTC'

UTC = UtcTzinfo()


def typename(obj):
  """Returns the type of obj as a string."""


  if hasattr(obj, '__class__'):
    return getattr(obj, '__class__').__name__
  else:
    return type(obj).__name__


def ValidateString(value,
                   name='unused',
                   exception=datastore_errors.BadValueError,
                   max_len=_MAX_STRING_LENGTH,
                   empty_ok=False):
  """Raises an exception if the value is not a valid string or a subclass thereof.

  A string is valid if it's not empty, no more than `_MAX_STRING_LENGTH` bytes,
  and not a Blob. The exception type can be specified with the exception
  argument; it defaults to `BadValueError`.

  Args:
    value: The value to validate.
    name: The name of this value; used in the exception message.
    exception: The type of exception to raise.
    max_len: The maximum allowed length, in bytes.
    empty_ok: Allow empty value.
  """
  if value is None and empty_ok:
    return
  if (
      not isinstance(value, (six.text_type, six.binary_type)) or
      isinstance(value, Blob)):
    raise exception('%s should be a string; received %s (a %s):' %
                    (name, value, typename(value)))
  if not value and not empty_ok:
    raise exception('%s must not be empty.' % name)




  conversion_kwargs = {}
  if six.PY3:
    conversion_kwargs = dict(errors='surrogatepass')
  if isinstance(value, six.text_type) and len(value.encode('utf-8', **conversion_kwargs)) > max_len:
    raise exception('%s must be under %d bytes.' % (name, max_len))
  if isinstance(value, str) and len(value) > max_len:
    raise exception('%s must be under %d bytes.' % (name, max_len))


def ValidateInteger(value,
                   name='unused',
                   exception=datastore_errors.BadValueError,
                   empty_ok=False,
                   zero_ok=False,
                   negative_ok=False):
  """Raises an exception if value is not a valid integer.

  An integer is valid if it's not negative or empty and is an integer
  (either `int` or `long`). The exception type raised can be specified
  with the exception argument; it defaults to `BadValueError`.

  Args:
    value: The value to validate.
    name: The name of this value; used in the exception message.
    exception: The type of exception to raise.
    empty_ok: Allow `None` value.
    zero_ok: Allow zero value.
    negative_ok: Allow negative value.
  """
  if value is None and empty_ok:
    return
  if not isinstance(value, six.integer_types):
    raise exception('%s should be an integer; received %s (a %s).' %
                    (name, value, typename(value)))
  if not value and not zero_ok:
    raise exception('%s must not be 0 (zero)' % name)
  if value < 0 and not negative_ok:
    raise exception('%s must not be negative.' % name)

def ResolveAppId(app):
  """Validates the application ID, providing a default.

  Args:
    app: The app ID argument value to be validated.

  Returns:
    The value of the app, or the substituted default. Always a non-empty string.

  Raises:
    `BadArgumentError` if the value is empty or not a string.
  """
  if app is None:
    app = full_app_id.get()
  ValidateString(app, 'app', datastore_errors.BadArgumentError)
  return app


def ResolveNamespace(namespace):
  """Validates the app namespace, providing a default.

  If the argument is `None`, `namespace_manager.get_namespace()` is substituted.

  Args:
    namespace: The namespace argument value to be validated.

  Returns:
    The value of namespace, or the substituted default. The empty string is used
    to denote the empty namespace.

  Raises:
    `BadArgumentError` if the value is not a string.
  """
  if namespace is None:
    namespace = namespace_manager.get_namespace()
  else:
    namespace_manager.validate_namespace(
        namespace, datastore_errors.BadArgumentError)
  return namespace


def EncodeAppIdNamespace(app_id, namespace):
  """Concatenates the application ID and namespace into a single string.

  This method is needed for XML and `datastore_file_stub`.

  Args:
    app_id: The application ID to encode.
    namespace: The namespace to encode.

  Returns:
    The string encoding for the `app_id`, namespace pair.
  """
  if not namespace:
    return app_id
  else:
    return app_id + _NAMESPACE_SEPARATOR + namespace


def DecodeAppIdNamespace(app_namespace_str):
  """Decodes `app_namespace_str` into an `(app_id, namespace)` pair.

  This method is the reverse of `EncodeAppIdNamespace` and is needed for
  `datastore_file_stub`.

  Args:
    app_namespace_str: An encoded `(app_id, namespace)` pair created by
      `EncodeAppIdNamespace`.

  Returns:
    `(app_id, namespace)` pair encoded in `app_namespace_str`.
  """
  sep = app_namespace_str.find(_NAMESPACE_SEPARATOR)
  if sep < 0:
    return (app_namespace_str, '')
  else:
    return (app_namespace_str[0:sep], app_namespace_str[sep + 1:])


def SetNamespace(proto, namespace):
  """Sets the namespace for a protocol buffer or clears the field.

  Args:
    proto: An `entity_pb2.Reference` to update
    namespace: The new namespace. `None` or an empty string will clear out the
        field.
  """
  if not namespace:
    proto.ClearField('name_space')
  else:
    proto.name_space = namespace


def PartitionString(value, separator):
  """Returns a 3-element tuple containing the part before the separator and the part after the separator.

  Equivalent to python2.5 `str.partition()`.

  Args:
    value: String to be partitioned.
    separator: Separator string.
  """
  index = value.find(separator)
  if index == -1:
    return (value, '', value[0:0])
  else:
    return (value[0:index], separator, value[index+len(separator):len(value)])


@cmp_compat.total_ordering_from_cmp
class Key(object):
  """The primary key for a datastore entity.

  A datastore `GUID`. A `Key` instance uniquely identifies an entity across all
  apps, and includes all information necessary to fetch the entity from the
  datastore with `Get()`.

  `Key` implements `__hash__`, and key instances are immutable, so Keys may be
  used in sets and as dictionary keys.
  """
  __reference = None

  def __init__(self, encoded=None):
    """Constructor. Creates a Key from a string.

    Args:
      # a base64-encoded primary key, generated by Key.__str__
      encoded: str
    """
    self._bytes = None
    if encoded is not None:
      if isinstance(encoded, bytes):
        pass
      elif isinstance(encoded, six.text_type):
        encoded = encoded.encode('utf-8')
      else:
        try:
          repr_encoded = repr(encoded)
        except:
          repr_encoded = "<couldn't encode>"
        raise datastore_errors.BadArgumentError(
          'Key() expects a string; received %s (a %s).' %
          (repr_encoded, typename(encoded)))
      try:

        modulo = len(encoded) % 4
        if modulo != 0:
          encoded += (b'=' * (4 - modulo))






        self._bytes = encoded
        encoded_pb = base64.urlsafe_b64decode(self._bytes)
        self.__reference = entity_pb2.Reference.FromString(encoded_pb)
        assert self.__reference.IsInitialized()


        self._bytes = self._bytes.rstrip(b'=')

      except (AssertionError, TypeError, binascii.Error) as e:
        raise datastore_errors.BadKeyError(
          'Invalid string key %s. Details: %s' % (encoded, e))
      except Exception as e:

        if (e.__class__.__name__ == 'ProtocolBufferDecodeError' or
            e.__class__.__name__ == 'DecodeError'):
          raise datastore_errors.BadKeyError('Invalid string key %s.' % encoded)
        else:
          raise
    else:

      self.__reference = entity_pb2.Reference()

  def to_path(self, _default_id=None, _decode=True, _fail=True):
    """Construct the "path" of this key as a list.

    Returns:
      A list [kind_1, id_or_name_1, ..., kind_n, id_or_name_n] of the key path.

    Raises:
      datastore_errors.BadKeyError if this key does not have a valid path.
    """







    path = []
    for path_element in self.__reference.path.element:
      path.append(path_element.type)
      if path_element.HasField('name'):
        path.append(path_element.name)
      elif path_element.HasField('id'):
        path.append(path_element.id)
      elif _default_id is not None:
        path.append(_default_id)
      else:
        raise datastore_errors.BadKeyError('Incomplete key found in to_path')
    return path

  @staticmethod
  def from_path(*args, **kwds):
    """Static method to construct a `Key` out of a "path" (`kind`, `id` or `name`, ...).

    This is useful when an application wants to use just the id or name portion
    of a key in, e.g., a URL, where the rest of the URL provides enough
    context to fill in the other properties. For example, the app id (always
    implicit), the entity kind, and possibly an ancestor key. Since ids and
    names are usually small, they're more attractive for use in end-user-visible
    URLs than the full string representation of a key.

    Args:
      kind: The entity kind (a str or unicode instance)
      id_or_name: The id (an int or long) or name (a str or unicode instance)
      parent: Optional parent `Key`; default `None`.
      namespace: Optional namespace to use otherwise namespace_manager's
        default namespace is used.

    Returns:
      A new `Key` instance whose `.kind()` and `.id()` or `.name()` methods
      return the *last* kind and id or name positional arguments passed.

    Raises:
      `BadArgumentError` for invalid arguments.
      `BadKeyError` if the parent key is incomplete.
    """

    parent = kwds.pop('parent', None)

    app_id = ResolveAppId(kwds.pop('_app', None))



    namespace = kwds.pop('namespace', None)


    if kwds:
      raise datastore_errors.BadArgumentError(
          'Excess keyword arguments ' + repr(kwds))


    if not args or len(args) % 2:
      raise datastore_errors.BadArgumentError(
          'A non-zero even number of positional arguments is required '
          '(kind, id or name, kind, id or name, ...); received %s' % repr(args))


    if parent is not None:
      if not isinstance(parent, Key):
        raise datastore_errors.BadArgumentError(
            'Expected None or a Key as parent; received %r (a %s).' %
            (parent, typename(parent)))
      if namespace is None:
        namespace = parent.namespace()
      if not parent.has_id_or_name():
        raise datastore_errors.BadKeyError(
            'The parent Key is incomplete.')
      if app_id != parent.app() or namespace != parent.namespace():
        raise datastore_errors.BadArgumentError(
            'The app/namespace arguments (%s/%s) should match '
            'parent.app/namespace() (%s/%s)' %
            (app_id, namespace, parent.app(), parent.namespace()))


    namespace = ResolveNamespace(namespace)


    key = Key()
    ref = key.__reference
    if parent is not None:
      ref.CopyFrom(parent.__reference)
    else:
      ref.app = app_id
      SetNamespace(ref, namespace)



    path = ref.path
    for i in range(0, len(args), 2):
      kind, id_or_name = args[i:i+2]
      if isinstance(kind, six.string_types):
        kind = kind.encode('utf-8')
      else:
        raise datastore_errors.BadArgumentError(
            'Expected a string kind as argument %d; received %r (a %s).' %
            (i + 1, kind, typename(kind)))
      elem = path.element.add()
      elem.type = kind
      if isinstance(id_or_name, six.integer_types):
        elem.id = id_or_name
      elif isinstance(id_or_name, six.string_types):
        ValidateString(id_or_name, 'name')
        elem.name = id_or_name.encode('utf-8')
      else:
        raise datastore_errors.BadArgumentError(
            'Expected an integer id or string name as argument %d; '
            'received %r (a %s).' % (i + 2, id_or_name, typename(id_or_name)))


    assert ref.IsInitialized()
    return key

  def app(self):
    """Returns this entity's app id, a string."""
    if self.__reference.app:
      return self.__reference.app
    else:
      return None

  def namespace(self):
    """Returns this entity's namespace, a string."""
    if self.__reference.HasField('name_space'):
      return self.__reference.name_space
    else:
      return ''



  def kind(self):
    """Returns this entity's kind, as a string."""
    if self.__reference.path.element:
      return self.__reference.path.element[-1].type
    else:
      return None

  def id(self):
    """Returns this entity's id, or None if it doesn't have one."""
    elems = self.__reference.path.element
    if elems and elems[-1].HasField('id') and elems[-1].id:
      return elems[-1].id
    else:
      return None

  def name(self):
    """Returns this entity's name, or None if it doesn't have one."""
    elems = self.__reference.path.element
    if elems and elems[-1].HasField('name') and elems[-1].name:
      return elems[-1].name
    else:
      return None

  def id_or_name(self):
    """Returns this entity's id or name, whichever it has, or None."""
    if self.id() is not None:
      return self.id()
    else:

      return self.name()

  def has_id_or_name(self):
    """Returns True if this entity has an id or name, False otherwise.
    """
    elems = self.__reference.path.element
    if elems:
      e = elems[-1]
      return bool(e.name or e.id)
    else:
      return False

  def parent(self):
    """Returns this entity's parent, as a Key. If this entity has no parent,
    returns None."""
    if len(self.__reference.path.element) > 1:
      parent = Key()
      parent.__reference.CopyFrom(self.__reference)
      del parent.__reference.path.element[-1]
      return parent
    else:
      return None

  def ToTagUri(self):
    """Returns a tag: URI for this entity for use in XML output.

    Foreign keys for entities may be represented in XML output as tag `URIs`.
    RFC 4151 describes the tag URI scheme. From http://taguri.org/:

      The tag algorithm lets people mint - create - identifiers that no one
      else using the same algorithm could ever mint. It is simple enough to do
      in your head, and the resulting identifiers can be easy to read, write,
      and remember. The identifiers conform to the URI (URL) Syntax.

    Tag URIs for entities use the app's auth domain and the date that the URI
     is generated. The namespace-specific part is `<kind>[<key>]`.

    For example, here is the tag URI for a Kitten with the key "Fluffy" in the
    catsinsinks app:

      tag:catsinsinks.googleapps.com,2006-08-29:Kitten[Fluffy]

    Raises a `BadKeyError` if this entity's key is incomplete.
    """
    if not self.has_id_or_name():
      raise datastore_errors.BadKeyError(
        'ToTagUri() called for an entity with an incomplete key.')

    return u'tag:%s.%s,%s:%s[%s]' % (

        saxutils.escape(EncodeAppIdNamespace(self.app(), self.namespace())),
        os.environ['AUTH_DOMAIN'],
        datetime.date.today().isoformat(),
        saxutils.escape(self.kind()),
        saxutils.escape(str(self)))

  ToXml = ToTagUri

  def entity_group(self):
    """Returns this key's entity group as a Key.

    Note that the returned Key will be incomplete if this Key is for a root
    entity and it is incomplete.
    """
    group = Key._FromPb(self.__reference)
    del group.__reference.path.element[1:]
    return group

  @staticmethod
  def _FromPb(pb):
    """Static factory method.

    Creates a Key from an entity_pb2.Reference.

    Not intended to be used by application developers. Enforced by hiding the
    entity_pb2 classes.

    Args:
      pb: entity_pb2.Reference

    Returns:
      A datastore_types.Key.
    """
    if not isinstance(pb, entity_pb2.Reference):
      raise datastore_errors.BadArgumentError(
          'Key constructor takes an entity_pb2.Reference; received %s (a %s).' %
          (pb, typename(pb)))

    key = Key()
    key.__reference = entity_pb2.Reference()
    key.__reference.CopyFrom(pb)
    return key

  def _ToPb(self):
    """Converts this Key to its protocol buffer representation.

    Not intended to be used by application developers. Enforced by hiding the
    entity_pb classes.e

    Returns:
      # The `Reference` `PB` representation of this Key
      entity_pb2.Reference
    """
    pb = entity_pb2.Reference()
    pb.CopyFrom(self.__reference)

    return pb

  def __str__(self):
    """Encodes this `Key` as an opaque string.

    Returns a string representation of this key, suitable for use in HTML,
    URLs, and other similar use cases. If the entity's key is incomplete,
    raises a `BadKeyError`.

    Unfortunately, this string encoding isn't particularly compact, and its
    length varies with the length of the path. If you want a shorter identifier
    and you know the kind and parent (if any) ahead of time, consider using just
    the entity's id or name.

    Returns:
      string
    """



    try:
      if self._bytes is not None:
        return self._bytes.decode('utf-8')
    except AttributeError:
      pass
    if (self.has_id_or_name()):
      encoded = base64.urlsafe_b64encode(self.__reference.SerializeToString())
      self._bytes = encoded.replace(b'=', b'')
    else:
      raise datastore_errors.BadKeyError(
        'Cannot string encode an incomplete key!\n%s' % self.__reference)
    return self._bytes.decode('utf-8')


  def __repr__(self):
    """Returns a Python string of the form `datastore_types.Key.from_path(...)` that can be used to recreate this key.

    Returns:
      String
    """
    args = []
    for elem in self.__reference.path.element:
      args.append(six.text_type(repr(elem.type)))
      if elem.HasField('name'):
        args.append(six.text_type(repr(elem.name)))
      else:
        args.append(repr(elem.id))

    args.append('_app=%r' % self.__reference.app)
    if self.__reference.HasField('name_space'):
      args.append('namespace=%r' % six.ensure_text(self.__reference.name_space))
    return u'datastore_types.Key.from_path(%s)' % ', '.join(args)

  def __cmp__(self, other):
    """Returns negative, zero, or positive when comparing two keys.

    TODO: for API v2, we should change this to make incomplete keys, ie
    keys without an id or name, not equal to any other keys.

    Args:
      other: Key to compare to.

    Returns:
      Negative if self is less than "other"
      Zero if "other" is equal to self
      Positive if self is greater than "other"
    """

    if not isinstance(other, Key):
      return -2

    self_args = [self.__reference.app, self.__reference.name_space]
    self_args += self.to_path(_default_id=0, _decode=False)

    other_args = [other.__reference.app, other.__reference.name_space]
    other_args += other.to_path(_default_id=0, _decode=False)

    for self_component, other_component in zip(self_args, other_args):
      comparison = cmp_compat.cmp(self_component, other_component)
      if comparison != 0:
        return comparison

    return cmp_compat.cmp(len(self_args), len(other_args))

  def __hash__(self):
    """Returns an integer hash of this key.

    Implements Python's hash protocol so that Keys may be used in sets and as
    dictionary keys.

    Returns:
      int
    """
    args = self.to_path(_default_id=0, _fail=False)
    args.append(self.__reference.app)
    return hash(type(args)) ^ hash(tuple(args))


class _OverflowDateTime(_PREFERRED_NUM_TYPE):
  """Container for GD_WHEN values that don't fit into a datetime.datetime.

  This class only exists to safely round-trip GD_WHEN values that are too large
  to fit in a datetime.datetime instance e.g. that were created by Java
  applications. It should not be created directly.
  """
  pass

def _EmptyList(val):
  if val is not None:
    raise datastore_errors.BadValueError('value should be None.')
  return []

def _When(val):
  """Coverts a GD_WHEN value to the appropriate type."""
  try:
    return _EPOCH + datetime.timedelta(microseconds=val)
  except OverflowError:
    return _OverflowDateTime(val)










class Category(six.text_type):
  """A tag, ie a descriptive word or phrase. Entities may be tagged by users,
  and later returned by a queries for that tag. Tags can also be used for
  ranking results (frequency), photo captions, clustering, activity, etc.

  This is the Atom `category` element. In XML output, the tag is provided as
  the term attribute. See:
  http://www.atomenabled.org/developers/syndication/#category

  Raises `BadValueError` if tag is not a string or subtype.
  """
  TERM = 'user-tag'

  def __init__(self, tag):
    super(Category, self).__init__()
    ValidateString(tag, 'tag')

  def ToXml(self):
    return u'<category term="%s" label=%s />' % (Category.TERM,
                                                 saxutils.quoteattr(self))


class Link(six.text_type):
  """A fully qualified URL. Usually http: scheme, but may also be file:, ftp:,
  news:, among others.

  If you have email (`mailto:`) or instant messaging (`aim:`, `xmpp:`) links,
  consider using the `Email` or `IM` classes instead.

  This is the Atom `link` element. In XML output, the link is provided as the
  href attribute. See:
  http://www.atomenabled.org/developers/syndication/#link

  Raises `BadValueError` if link is not a fully qualified, well-formed URL.
  """
  def __init__(self, link):
    super(Link, self).__init__()
    ValidateString(link, 'link', max_len=_MAX_LINK_PROPERTY_LENGTH)

    scheme, domain, path, _, _, _ = urllib.parse.urlparse(link)
    if (not scheme or (scheme != 'file' and not domain) or
                      (scheme == 'file' and not path)):
      raise datastore_errors.BadValueError('Invalid URL: %s' % link)

  def ToXml(self):
    return u'<link href=%s />' % saxutils.quoteattr(self)


class Email(six.text_type):
  """An RFC2822 email address. Makes no attempt at validation.

  This is the `gd:email` element. In XML output, the email address is provided
  as the address attribute. See:
  https://developers.google.com/gdata/docs/1.0/elements#gdEmail

  Raises `BadValueError` if email is not a valid email address.
  """
  def __init__(self, email):
    super(Email, self).__init__()
    ValidateString(email, 'email')

  def ToXml(self):
    return u'<gd:email address=%s />' % saxutils.quoteattr(self)


@cmp_compat.total_ordering_from_cmp
class GeoPt(object):
  """A geographical point, specified by floating-point latitude and longitude
  coordinates. Often used to integrate with mapping sites like Google Maps.

  This is the `georss:point` element. In XML output, the coordinates are
  provided as the `lat` and `lon` attributes. See: http://georss.org/

  Serializes to `<lat>,<lon>`. Raises `BadValueError` if it's passed an invalid
  serialized string, or if `lat` and `lon` are not valid floating points in the
  ranges [-90, 90] and [-180, 180], respectively.
  """
  lat = None
  lon = None

  def __init__(self, lat, lon=None):
    if lon is None:

      try:
        split = lat.split(',')
        lat, lon = split
      except (AttributeError, ValueError):
        raise datastore_errors.BadValueError(
          'Expected a "lat,long" formatted string; received %s (a %s).' %
          (lat, typename(lat)))

    try:
      lat = float(lat)
      lon = float(lon)
      if abs(lat) > 90:
        raise datastore_errors.BadValueError(
          'Latitude must be between -90 and 90; received %f' % lat)
      if abs(lon) > 180:
        raise datastore_errors.BadValueError(
          'Longitude must be between -180 and 180; received %f' % lon)
    except (TypeError, ValueError):

      raise datastore_errors.BadValueError(
        'Expected floats for lat and long; received %s (a %s) and %s (a %s).' %
        (lat, typename(lat), lon, typename(lon)))

    self.lat = lat
    self.lon = lon

  def __cmp__(self, other):
    """Returns negative, zero, or positive when comparing two `GeoPts`."""

    if not isinstance(other, GeoPt):
      try:
        other = GeoPt(other)
      except datastore_errors.BadValueError:
        return NotImplemented


    lat_cmp = cmp_compat.cmp(self.lat, other.lat)
    if lat_cmp != 0:
      return lat_cmp
    else:
      return cmp_compat.cmp(self.lon, other.lon)

  def __hash__(self):
    """Returns an integer hash of this point.

    Implements Python's hash protocol so that `GeoPts` may be used in sets and
    as dictionary keys.

    Returns:
      int
    """
    return hash((self.lat, self.lon))

  def __repr__(self):
    """Returns a string of the form `datastore_types.GeoPt([lat], [lon])`.

    Returns:
      String
    """
    return 'datastore_types.GeoPt(%r, %r)' % (self.lat, self.lon)

  def __unicode__(self):
    return u'%s,%s' % (six.text_type(self.lat),
                       six.text_type(self.lon))

  __str__ = __unicode__

  def ToXml(self):
    return u'<georss:point>%s %s</georss:point>' % (six.text_type(
        self.lat), six.text_type(self.lon))


@cmp_compat.total_ordering_from_cmp
class IM(object):
  """An instant messaging handle. Includes both an address and its protocol.
  The protocol value is either a standard IM scheme or a URL identifying the
  IM network for the protocol. Possible values include:

    Value                           Description
    sip                             SIP/SIMPLE
    unknown                         Unknown or unspecified
    xmpp                            XMPP/Jabber
    http://aim.com/                 AIM
    http://icq.com/                 ICQ
    http://talk.google.com/         Google Talk
    http://messenger.msn.com/       MSN Messenger
    http://messenger.yahoo.com/     Yahoo Messenger
    http://sametime.com/            Lotus Sametime
    http://gadu-gadu.pl/            Gadu-Gadu

  This is the gd:im element. In XML output, the address and protocol are
  provided as the address and protocol attributes, respectively. See:
  https://developers.google.com/gdata/docs/1.0/elements#gdIm

  Serializes to `<protocol> <address>`. Raises `BadValueError` if tag is not a
  standard IM scheme or a URL.
  """
  PROTOCOLS = [ 'sip', 'unknown', 'xmpp' ]

  protocol = None
  address = None

  def __init__(self, protocol, address=None):
    if address is None:

      try:
        split = protocol.split(' ', 1)
        protocol, address = split
      except (AttributeError, ValueError):
        raise datastore_errors.BadValueError(
          'Expected string of format "protocol address"; received %s' %
          (protocol,))

    ValidateString(address, 'address')
    if protocol not in self.PROTOCOLS:

      Link(protocol)

    self.address = address
    self.protocol = protocol

  def __cmp__(self, other):
    """Returns negative, zero, or positive when comparing two `IM`s."""

    if not isinstance(other, IM):
      try:
        other = IM(other)
      except datastore_errors.BadValueError:
        return NotImplemented











    return cmp_compat.cmp((self.address, self.protocol),
                          (other.address, other.protocol))

  def __repr__(self):
    """Returns a string of the form `datastore_types.IM('address', 'protocol')`.

    Returns:
      string
    """
    return 'datastore_types.IM(%r, %r)' % (self.protocol, self.address)

  def __unicode__(self):
    return u'%s %s' % (self.protocol, self.address)

  __str__ = __unicode__

  def ToXml(self):
    return (u'<gd:im protocol=%s address=%s />' %
            (saxutils.quoteattr(self.protocol),
             saxutils.quoteattr(self.address)))

  def __len__(self):
    return len(six.text_type(self))


class PhoneNumber(six.text_type):
  """A human-readable phone number or address.

  No validation is performed. Phone numbers have many different formats -
  local, long distance, domestic, international, internal extension, TTY,
  VOIP, SMS, and alternative networks like Skype, XFire and Roger Wilco. They
  all have their own numbering and addressing formats.

  This is the `gd:phoneNumber` element. In XML output, the phone number is
  provided as the text of the element. See:
  https://developers.google.com/gdata/docs/1.0/elements#gdPhoneNumber

  Raises `BadValueError` if phone is not a string or subtype.
  """
  def __init__(self, phone):
    super(PhoneNumber, self).__init__()
    ValidateString(phone, 'phone')

  def ToXml(self):
    return u'<gd:phoneNumber>%s</gd:phoneNumber>' % saxutils.escape(self)


class PostalAddress(six.text_type):
  """A human-readable mailing address. Again, mailing address formats vary
  widely, so no validation is performed.

  This is the gd:postalAddress element. In XML output, the address is provided
  as the text of the element. See:
  https://developers.google.com/gdata/docs/1.0/elements#gdPostalAddress

  Raises `BadValueError` if address is not a string or subtype.
  """
  def __init__(self, address):
    super(PostalAddress, self).__init__()
    ValidateString(address, 'address')

  def ToXml(self):
    return u'<gd:postalAddress>%s</gd:postalAddress>' % saxutils.escape(self)


class Rating(_PREFERRED_NUM_TYPE):
  """A user-provided integer rating for a piece of content. Normalized to a
  0-100 scale.

  This is the gd:rating element. In XML output, the address is provided
  as the text of the element. See:
  https://developers.google.com/gdata/docs/1.0/elements#gdRating

  Serializes to the decimal string representation of the rating. Raises
  `BadValueError` if the rating is not an integer in the range [0, 100].
  """
  MIN = 0
  MAX = 100

  def __init__(self, rating):
    super(Rating, self).__init__()
    if isinstance(rating, float) or isinstance(rating, complex):

      raise datastore_errors.BadValueError(
        'Expected int or long; received %s (a %s).' %
        (rating, typename(rating)))

    try:
      if (_PREFERRED_NUM_TYPE(rating) < Rating.MIN
          or _PREFERRED_NUM_TYPE(rating) > Rating.MAX):
        raise datastore_errors.BadValueError()
    except ValueError:

      raise datastore_errors.BadValueError(
          'Expected int or long; received %s (a %s).' %
          (rating, typename(rating)))

  def ToXml(self):
    return (u'<gd:rating value="%d" min="%d" max="%d" />' %
            (self, Rating.MIN, Rating.MAX))


class Text(six.text_type):
  """A long string type.

  Strings of any length can be stored in the datastore using this
  type. It behaves identically to the Python unicode type, except for
  the constructor, which only accepts str and unicode arguments.
  """

  def __new__(cls, arg=None, encoding=None):
    """Constructor.

    We only accept unicode and str instances, the latter with encoding.

    Args:
      arg: optional unicode or str instance; default u''
      encoding: optional encoding; disallowed when `isinstance(arg, unicode)`,
                defaults to `ascii` when `isinstance(arg, str)`;
    """
    if arg is None:
      arg = u''
    if isinstance(arg, six.text_type):
      if encoding is not None:
        raise TypeError('Text() with a unicode argument '
                        'should not specify an encoding')
      return super(Text, cls).__new__(cls, arg)

    if isinstance(arg, bytes):
      if encoding is None:
        encoding = 'ascii'
      return super(Text, cls).__new__(cls, arg, encoding)

    raise TypeError('Text() argument should be str or unicode, not %s' %
                    type(arg).__name__)


class _BaseByteType(bytes):
  """A base class for datastore types that are encoded as bytes.

  This behaves identically to the Python bytes type, except for the
  constructor, which only accepts bytes arguments.
  """

  def __new__(cls, arg=None):
    """Constructor.

    We only accept bytes instances.

    Args:
      arg: optional bytes instance (default b'')
    """
    if arg is None:
      arg = b''

    if isinstance(arg, bytes):
      return super(_BaseByteType, cls).__new__(cls, arg)

    raise TypeError('%s() argument should be bytes instance, not %s' %
                    (cls.__name__, type(arg).__name__))

  def ToXml(self):
    """Output bytes as XML.

    Returns:
      Base64 encoded version of itself for safe insertion in to an XML document.
    """
    encoded = base64.urlsafe_b64encode(self).decode('utf-8')
    return saxutils.escape(encoded)

  if six.PY3:


    def __str__(self):
      return self.decode('utf-8')


class Blob(_BaseByteType):
  """A blob type, appropriate for storing binary data of any length.

  This behaves identically to the Python bytes type, except for the
  constructor, which only accepts bytes arguments.
  """

  def __new__(cls, *args, **kwargs):
    self = super(Blob, cls).__new__(cls, *args, **kwargs)
    self._meaning_uri = None
    return self

  @property
  def meaning_uri(self):
    return self._meaning_uri

  @meaning_uri.setter
  def meaning_uri(self, value):
    self._meaning_uri = value


class EmbeddedEntity(_BaseByteType):
  """A proto encoded `EntityProto`.

  This behaves identically to Blob, except for the
  constructor, which accepts a bytes or `EntityProto` argument.

  Can be decoded using `datastore.Entity.FromProto()`,
  `db.model_from_protobuf()`, or
  `ndb.LocalStructuredProperty`.
  """

  def __new__(cls, arg=None):
    """Constructor.

    Args:
      arg: Optional str or `EntityProto` instance (default `''`).
    """
    if isinstance(arg, entity_pb2.EntityProto):
      arg = arg.SerializePartialToString()
    return super(EmbeddedEntity, cls).__new__(cls, arg)


class ByteString(_BaseByteType):
  """A byte-string type, appropriate for storing short amounts of indexed data.

  This behaves identically to Blob, except it's used only for short, indexed
  byte strings.
  """
  pass


@cmp_compat.total_ordering_from_cmp
class BlobKey(object):
  """Key used to identify a blob in Blobstore.

  This object wraps a string that gets used internally by the Blobstore API
  to identify application blobs. The `BlobKey` corresponds to the entity name
  of the underlying `BlobReference` entity.

  This class is exposed in the API in both `google.appengine.ext.db` and
  `google.appengine.ext.blobstore`.
  """

  def __init__(self, blob_key):
    """Constructor.

    Used to convert a string to a `BlobKey`.  Normally used internally by
    Blobstore API.

    Args:
      blob_key:  Key name of `BlobReference` that this key belongs to.
    """
    ValidateString(blob_key, 'blob-key', empty_ok=True)
    self.__blob_key = blob_key

  def __str__(self):
    """Convert to string."""
    return six.ensure_str(self.__blob_key)

  def __repr__(self):
    """Returns a Python string of the form 'datastore_types.BlobKey(...)' that can be used to recreate this key.

    Returns:
      String
    """
    return 'datastore_types.%s(%r)' % (type(self).__name__, self.__blob_key)

  def __cmp__(self, other):


    if type(other) is type(self):
      return cmp_compat.cmp(str(self), str(other))
    elif isinstance(other, six.string_types):
      return cmp_compat.cmp(self.__blob_key, other)
    else:
      return NotImplemented

  def __hash__(self):
    return hash(self.__blob_key)

  def ToXml(self):
    return str(self)



_PROPERTY_MEANINGS = {



    Blob: entity_pb2.Property.BLOB,
    EmbeddedEntity: entity_pb2.Property.ENTITY_PROTO,
    ByteString: entity_pb2.Property.BYTESTRING,
    Text: entity_pb2.Property.TEXT,
    datetime.datetime: entity_pb2.Property.GD_WHEN,
    datetime.date: entity_pb2.Property.GD_WHEN,
    datetime.time: entity_pb2.Property.GD_WHEN,
    _OverflowDateTime: entity_pb2.Property.GD_WHEN,
    Category: entity_pb2.Property.ATOM_CATEGORY,
    Link: entity_pb2.Property.ATOM_LINK,
    Email: entity_pb2.Property.GD_EMAIL,
    GeoPt: entity_pb2.Property.GEORSS_POINT,
    IM: entity_pb2.Property.GD_IM,
    PhoneNumber: entity_pb2.Property.GD_PHONENUMBER,
    PostalAddress: entity_pb2.Property.GD_POSTALADDRESS,
    Rating: entity_pb2.Property.GD_RATING,
    BlobKey: entity_pb2.Property.BLOBKEY,
}


_PROPERTY_TYPES = frozenset([
    Blob,
    EmbeddedEntity,
    ByteString,
    bool,
    Category,
    datetime.datetime,
    _OverflowDateTime,
    Email,
    float,
    GeoPt,
    IM,
    int,
    Key,
    Link,
    _PREFERRED_NUM_TYPE,
    PhoneNumber,
    PostalAddress,
    Rating,
    str,
    Text,
    type(None),
    six.text_type,
    users.User,
    BlobKey,
    bytes,
])




_RAW_PROPERTY_TYPES = (Blob, Text, EmbeddedEntity)
_RAW_PROPERTY_MEANINGS = (entity_pb2.Property.BLOB, entity_pb2.Property.TEXT,
                          entity_pb2.Property.ENTITY_PROTO)


def ValidatePropertyInteger(name, value):
  """Raises an exception if the supplied integer is invalid.

  Args:
    name: Name of the property this is for.
    value: Integer value.

  Raises:
    `OverflowError` if the value does not fit within a signed `int64`.
  """
  if not (-0x8000000000000000 <= value <= 0x7fffffffffffffff):
    raise OverflowError('%d is out of bounds for int64' % value)


def ValidateStringLength(name, value, max_len):
  """Raises an exception if the supplied string is too long.

  Args:
    name: Name of the property this is for.
    value: String value.
    max_len: Maximum length the string may be.

  Raises:
    `OverflowError` if the value is larger than the maximum length.
  """

  if isinstance(value, six.text_type):
    value = value.encode('utf-8')

  if len(value) > max_len:
    raise datastore_errors.BadValueError(
      'Property %s is %d bytes long; it must be %d or less. '
      'Consider Text instead, which can store strings of any length.' %
      (name, len(value), max_len))


def ValidatePropertyString(name, value):
  """Validates the length of an indexed string property.

  Args:
    name: Name of the property this is for.
    value: String value.
  """
  ValidateStringLength(name, value, max_len=_MAX_STRING_LENGTH)


def ValidatePropertyLink(name, value):
  """Validates the length of an indexed `Link` property.

  Args:
    name: Name of the property this is for.
    value: String value.
  """
  ValidateStringLength(name, value, max_len=_MAX_LINK_PROPERTY_LENGTH)


def ValidatePropertyNothing(name, value):
  """No-op validation function.

  Args:
    name: Name of the property this is for.
    value: Not used.
  """
  pass


def ValidatePropertyKey(name, value):
  """Raises an exception if the supplied `datastore.Key` instance is invalid.

  Args:
    name: Name of the property this is for.
    value: A `datastore.Key` instance.

  Raises:
    `datastore_errors.BadValueError` if the value is invalid.
  """
  if not value.has_id_or_name():
    raise datastore_errors.BadValueError(
        'Incomplete key found for reference property %s.' % name)







_VALIDATE_PROPERTY_VALUES = {
    Blob: ValidatePropertyNothing,
    EmbeddedEntity: ValidatePropertyNothing,
    ByteString: ValidatePropertyNothing,
    bool: ValidatePropertyNothing,
    Category: ValidatePropertyNothing,
    datetime.datetime: ValidatePropertyNothing,
    _OverflowDateTime: ValidatePropertyInteger,
    Email: ValidatePropertyNothing,
    float: ValidatePropertyNothing,
    GeoPt: ValidatePropertyNothing,
    IM: ValidatePropertyNothing,
    int: ValidatePropertyInteger,
    Key: ValidatePropertyKey,
    Link: ValidatePropertyNothing,
    _PREFERRED_NUM_TYPE: ValidatePropertyInteger,
    PhoneNumber: ValidatePropertyNothing,
    PostalAddress: ValidatePropertyNothing,
    Rating: ValidatePropertyInteger,
    str: ValidatePropertyNothing,
    Text: ValidatePropertyNothing,
    type(None): ValidatePropertyNothing,
    six.text_type: ValidatePropertyNothing,
    bytes: ValidatePropertyNothing,
    users.User: ValidatePropertyNothing,
    BlobKey: ValidatePropertyNothing,
}










_PROPERTY_TYPE_TO_INDEX_VALUE_TYPE = {
    six.text_type: bytes,
    Blob: bytes,
    EmbeddedEntity: bytes,
    ByteString: bytes,
    bool: bool,
    Category: bytes,
    datetime.datetime: _PREFERRED_NUM_TYPE,
    datetime.date: _PREFERRED_NUM_TYPE,
    datetime.time: _PREFERRED_NUM_TYPE,
    _OverflowDateTime: _PREFERRED_NUM_TYPE,
    Email: six.binary_type,
    float: float,
    GeoPt: GeoPt,
    IM: six.binary_type,
    int: _PREFERRED_NUM_TYPE,
    Key: Key,
    Link: six.binary_type,
    _PREFERRED_NUM_TYPE: _PREFERRED_NUM_TYPE,
    PhoneNumber: six.binary_type,
    PostalAddress: six.binary_type,
    Rating: _PREFERRED_NUM_TYPE,
    bytes: bytes,
    Text: bytes,
    type(None): type(None),
    users.User: users.User,
    BlobKey: bytes,
}
if six.PY2:


  _PROPERTY_TYPE_TO_INDEX_VALUE_TYPE[basestring] = bytes


assert set(_VALIDATE_PROPERTY_VALUES.keys()) == _PROPERTY_TYPES


def ValidateProperty(name, values, read_only=False):
  """Helper function for validating property values.

  Args:
    name: Name of the property this is for.
    value: Value for the property as a Python native type.
    read_only: Deprecated.

  Raises:
    `BadPropertyError` if the property name is invalid. `BadValueError` if the
    property did not validate correctly or the value was an empty list. Other
    exception types (like `OverflowError`) if the property value does not meet
    type-specific criteria.
  """
  ValidateString(name, 'property name', datastore_errors.BadPropertyError)

  values_type = type(values)


  if values_type is tuple:
    raise datastore_errors.BadValueError(
        'May not use tuple property value; property %s is %s.' %
        (name, repr(values)))


  if values_type is not list:
    values = [values]



  try:
    for v in values:
      prop_validator = _VALIDATE_PROPERTY_VALUES.get(v.__class__)
      if prop_validator is None:
        raise datastore_errors.BadValueError(
          'Unsupported type for property %s: %s' % (name, v.__class__))
      prop_validator(name, v)

  except (KeyError, ValueError, TypeError, IndexError, AttributeError) as msg:
    raise datastore_errors.BadValueError(
      'Error type checking values for property %s: %s' % (name, msg))





ValidateReadProperty = ValidateProperty



def PackBlob(name, value, pbvalue):
  """Packs a Blob property into a `entity_pb2.PropertyValue`.

  Args:
    name: The name of the property as a string.
    value: A Blob instance.
    pbvalue: The `entity_pbs.PropertyValue` to pack this value into.
  """
  pbvalue.stringValue = value


def PackString(name, value, pbvalue):
  """Packs a string-typed property into a entity_pb2.PropertyValue.

  Args:
    name: The name of the property as a string.
    value: A string, unicode, or string-like value instance.
    pbvalue: The entity_pb2.PropertyValue to pack this value into.
  """
  if isinstance(value, bytes):



    value.decode('ascii')
    pbvalue.stringValue = value
  else:
    pbvalue.stringValue = six.text_type(value).encode('utf-8')


def PackDatetime(name, value, pbvalue):
  """Packs a `datetime`-typed property into a `entity_pb2.PropertyValue`.

  Args:
    name: The name of the property as a string.
    value: A `datetime.datetime` instance.
    pbvalue: The `entity_pb2.PropertyValue` to pack this value into.
  """
  pbvalue.int64Value = DatetimeToTimestamp(value)


def DatetimeToTimestamp(value):
  """Converts a `datetime.datetime` to microseconds since the epoch, as a float.

  Args:
    value: `datetime.datetime`

  Returns:
    Value as a `long`.
  """
  if value.tzinfo:

    value = value.astimezone(UTC)
  return _PREFERRED_NUM_TYPE(
      calendar.timegm(value.timetuple()) * 1000000) + value.microsecond


def PackGeoPt(name, value, pbvalue):
  """Packs a `GeoPt` property into a `entity_pb2.PropertyValue`.

  Args:
    name: The name of the property as a string.
    value: A `GeoPt` instance.
    pbvalue: The `entity_pb2.PropertyValue` to pack this value into.
  """
  pbvalue.pointvalue.x = value.lat
  pbvalue.pointvalue.y = value.lon


def PackUser(name, value, pbvalue):
  """Packs a `User` property into a `entity_pb2.PropertyValue`.

  Args:
    name: The name of the property as a string.
    value: A `users.User` instance.
    pbvalue: The `entity_pb2.PropertyValue` to pack this value into.
  """
  pbvalue.uservalue.email = value.email().encode('utf-8')
  pbvalue.uservalue.auth_domain = value.auth_domain().encode('utf-8')
  pbvalue.uservalue.gaiaid = 0




  if value.user_id() is not None:
    pbvalue.uservalue.obfuscated_gaiaid = value.user_id().encode('utf-8')

  if value.federated_identity() is not None:
    pbvalue.uservalue.federated_identity = value.federated_identity().encode(
        'utf-8')

  if value.federated_provider() is not None:
    pbvalue.uservalue.federated_provider = value.federated_provider().encode(
        'utf-8')


def PackKey(name, value, pbvalue):
  """Packs a reference property into an `entity_pb2.PropertyValue`.

  Args:
    name: The name of the property as a string.
    value: A `Key` instance.
    pbvalue: The `entity_pb2.PropertyValue` to pack this value into.
  """
  ref = value._Key__reference
  pbvalue.referencevalue.app = ref.app
  SetNamespace(pbvalue.referencevalue, ref.name_space)
  for elem in ref.path.element:
    elementCopy = pbvalue.referencevalue.pathelement.add()
    datastore_pbs.copy_path_element(elem, elementCopy)


def PackBool(name, value, pbvalue):
  """Packs a boolean property into an `entity_pb2.PropertyValue`.

  Args:
    name: The name of the property as a string.
    value: A boolean instance.
    pbvalue: The `entity_pb2.PropertyValue` to pack this value into.
  """
  pbvalue.booleanValue = value


def PackInteger(name, value, pbvalue):
  """Packs an integer property into an `entity_pb2.PropertyValue`.

  Args:
    name: The name of the property as a string.
    value: An int or long instance.
    pbvalue: The `entity_pb2.PropertyValue` to pack this value into.
  """
  pbvalue.int64Value = value


def PackFloat(name, value, pbvalue):
  """Packs a float property into a `entity_pb2.PropertyValue`.

  Args:
    name: The name of the property as a string.
    value: A float instance.
    pbvalue: The `entity_pb2.PropertyValue` to pack this value into.
  """
  pbvalue.doubleValue = value







_PACK_PROPERTY_VALUES = {
    Blob: PackBlob,
    EmbeddedEntity: PackBlob,
    ByteString: PackBlob,
    bool: PackBool,
    Category: PackString,
    datetime.datetime: PackDatetime,
    _OverflowDateTime: PackInteger,
    Email: PackString,
    float: PackFloat,
    GeoPt: PackGeoPt,
    IM: PackString,
    int: PackInteger,
    Key: PackKey,
    Link: PackString,
    _PREFERRED_NUM_TYPE: PackInteger,
    PhoneNumber: PackString,
    PostalAddress: PackString,
    Rating: PackInteger,
    str: PackString,
    Text: PackString,
    type(None): lambda name, value, pbvalue: pbvalue.ClearField('stringValue'),
    six.text_type: PackString,
    users.User: PackUser,
    BlobKey: PackString,
    bytes: PackString,
}


assert set(_PACK_PROPERTY_VALUES.keys()) == _PROPERTY_TYPES


def ToPropertyPb(name, values):
  """Creates type-specific `entity_pb2.PropertyValues`.

  Determines the type and meaning of the `PropertyValue` based on the Python
  type of the input value(s).

  NOTE: This function does not validate anything!

  Args:
    name: String or unicode; the property name.
    values: The values for this property, either a single one or a list of them.
      All values must be a supported type. Lists of values must all be of the
      same type.

  Returns:
    A list of `entity_pb2.Property` instances.
  """
  encoded_name = six.ensure_str(name)

  values_type = type(values)
  if values_type is list and len(values) == 0:

    pb = entity_pb2.Property()
    pb.meaning = entity_pb2.Property.EMPTY_LIST
    pb.name = encoded_name
    pb.multiple = False

    pb.value.ClearField('stringValue')
    return pb
  elif values_type is list:
    multiple = True
  else:
    multiple = False
    values = [values]

  pbs = []
  for v in values:
    pb = entity_pb2.Property()
    pb.name = encoded_name
    pb.multiple = multiple

    meaning = _PROPERTY_MEANINGS.get(v.__class__)
    if meaning is not None:
      pb.meaning = meaning

    if hasattr(v, 'meaning_uri') and v.meaning_uri:
      pb.meaning_uri = v.meaning_uri

    pack_prop = _PACK_PROPERTY_VALUES[v.__class__]
    pack_prop(name, v, pb.value)
    pbs.append(pb)

  if multiple:
    return pbs
  else:
    return pbs[0]


def FromReferenceProperty(value):
  """Converts a reference `PropertyValue` to a `Key`.

  Args:
    value: `entity_pb2.PropertyValue`

  Returns:
    `Key`

  Raises:
    `BadValueError` if the value is not a `PropertyValue`.
  """
  assert isinstance(value, entity_pb2.PropertyValue)
  assert value.HasField('referencevalue')
  ref = value.referencevalue

  key = Key()
  key_ref = key._Key__reference
  key_ref.app = ref.app
  SetNamespace(key_ref, ref.name_space)

  for pathelem in ref.pathelement:
    element = key_ref.path.element.add()
    datastore_pbs.copy_path_element(pathelem, element)

  return key









_PROPERTY_CONVERSIONS = {
    entity_pb2.Property.GD_WHEN: _When,
    entity_pb2.Property.ATOM_CATEGORY: Category,
    entity_pb2.Property.ATOM_LINK: Link,
    entity_pb2.Property.GD_EMAIL: Email,
    entity_pb2.Property.GD_IM: IM,
    entity_pb2.Property.GD_PHONENUMBER: PhoneNumber,
    entity_pb2.Property.GD_POSTALADDRESS: PostalAddress,
    entity_pb2.Property.GD_RATING: Rating,
    entity_pb2.Property.BLOB: Blob,
    entity_pb2.Property.ENTITY_PROTO: EmbeddedEntity,
    entity_pb2.Property.BYTESTRING: ByteString,
    entity_pb2.Property.TEXT: Text,
    entity_pb2.Property.BLOBKEY: BlobKey,
    entity_pb2.Property.EMPTY_LIST: _EmptyList,
}

_NON_UTF8_MEANINGS = frozenset(
    (entity_pb2.Property.BLOB, entity_pb2.Property.ENTITY_PROTO,
     entity_pb2.Property.BYTESTRING, entity_pb2.Property.INDEX_VALUE))


def FromPropertyPb(pb):
  """Converts a property `PB` to a python value.

  Args:
    pb: `entity_pb2.Property`

  Returns:
    The return type is determined by the type of the argument, such as
    `string`, `int`, `bool`, `double`, `users.User`, or one of the `atom` or
    `gd` types.
  """
  pbval = pb.value
  meaning = pb.meaning

  if pbval.HasField('stringValue'):
    value = pbval.stringValue
    if not pb.HasField('meaning') or meaning not in _NON_UTF8_MEANINGS:

      value = value.decode('utf-8')
  elif pbval.HasField('int64Value'):


    value = _PREFERRED_NUM_TYPE(pbval.int64Value)
  elif pbval.HasField('booleanValue'):


    value = bool(pbval.booleanValue)
  elif pbval.HasField('doubleValue'):
    value = pbval.doubleValue
  elif pbval.HasField('referencevalue'):
    value = FromReferenceProperty(pbval)
  elif pbval.HasField('pointvalue'):
    value = GeoPt(pbval.pointvalue.x, pbval.pointvalue.y)
  elif pbval.HasField('uservalue'):
    email = pbval.uservalue.email
    auth_domain = pbval.uservalue.auth_domain
    obfuscated_gaiaid = pbval.uservalue.obfuscated_gaiaid

    federated_identity = None
    if pbval.uservalue.HasField('federated_identity'):
      federated_identity = pbval.uservalue.federated_identity



    value = users.User(email=email,
                       _auth_domain=auth_domain,
                       _user_id=obfuscated_gaiaid,
                       federated_identity=federated_identity,
                       _strict_mode=False)
  else:
    value = None

  try:
    if pb.HasField('meaning') and meaning in _PROPERTY_CONVERSIONS:
      conversion = _PROPERTY_CONVERSIONS[meaning]
      value = conversion(value)
      if (meaning == entity_pb2.Property.BLOB and pb.HasField('meaning_uri')):
        value.meaning_uri = pb.meaning_uri
  except (KeyError, ValueError, IndexError, TypeError, AttributeError) as msg:
    raise datastore_errors.BadValueError(
      'Error converting pb: %s\nException was: %s' % (pb, msg))

  return value


def RestoreFromIndexValue(index_value, data_type):
  """Restores an index value to the correct datastore type.

  Projection queries return property values directly from a datastore index.
  These values are the native datastore values that can be one of the following:
  `str`, `bool`, `long`, `float`, `GeoPt`, `Key`, or `User`. This function
  restores the original value when the original type is known.

  This function returns the value type returned when decoding a normal entity,
  not necessarily of type `data_type`. For example, `data_type=int` returns a
  `long` instance.

  Args:
    index_value: The value returned by `FromPropertyPb` for the projected
      property.
    data_type: The type of the value originally given to `ToPropertyPb`.

  Returns:
    The restored property value.

  Raises:
    `datastore_errors.BadValueError` if the value cannot be restored.
  """
  raw_type = _PROPERTY_TYPE_TO_INDEX_VALUE_TYPE.get(data_type)
  if raw_type is None:
    raise datastore_errors.BadValueError(
        'Unsupported data type (%r)' % data_type)

  if index_value is None:
    return index_value



  if not isinstance(index_value, raw_type):
    raise datastore_errors.BadValueError(
        'Unsupported conversion. Expected %r got %r' %
        (type(index_value), raw_type))

  meaning = _PROPERTY_MEANINGS.get(data_type)


  if isinstance(index_value, bytes) and meaning not in _NON_UTF8_MEANINGS:
    index_value = six.text_type(index_value, 'utf-8')


  conv = _PROPERTY_CONVERSIONS.get(meaning)
  if not conv:
    return index_value

  try:
    value = conv(index_value)
  except (KeyError, ValueError, IndexError, TypeError, AttributeError) as msg:
    raise datastore_errors.BadValueError(
      'Error converting value: %r\nException was: %s' % (index_value, msg))
  return value


def PropertyTypeName(value):
  """Returns the type name of the given property value, as a string.

  Raises `BadValueError` if the value is not a valid property type.

  Args:
    value: Any valid property value.

  Returns:
    String.
  """
  if value.__class__ in _PROPERTY_MEANINGS:
    meaning = _PROPERTY_MEANINGS[value.__class__]
    name = entity_pb2.Property.Meaning.DESCRIPTOR.values_by_number[meaning].name
    return name.lower().replace('_', ':')
  elif isinstance(value, six.string_types):
    return 'string'
  elif isinstance(value, users.User):
    return 'user'
  elif isinstance(value, bool):
    return 'bool'
  elif isinstance(value, _PREFERRED_NUM_TYPE):
    return 'int'
  elif value is None:
    return 'null'
  else:
    return typename(value).lower()


_PROPERTY_TYPE_STRINGS = {
    'string': six.text_type,
    'bool': bool,
    'int': _PREFERRED_NUM_TYPE,
    'null': type(None),
    'float': float,
    'key': Key,
    'blob': Blob,
    'entity:proto': EmbeddedEntity,
    'bytestring': ByteString,
    'text': Text,
    'user': users.User,
    'atom:category': Category,
    'atom:link': Link,
    'gd:email': Email,
    'gd:when': datetime.datetime,
    'georss:point': GeoPt,
    'gd:im': IM,
    'gd:phonenumber': PhoneNumber,
    'gd:postaladdress': PostalAddress,
    'gd:rating': Rating,
    'blobkey': BlobKey,
}


def FromPropertyTypeName(type_name):
  """Returns the python type given a type name.

  Args:
    type_name: A string representation of a datastore type name.

  Returns:
    A Python type.
  """
  return _PROPERTY_TYPE_STRINGS[type_name]


def PropertyValueFromString(type_,
                            value_string,
                            _auth_domain=None):
  """Returns an instance of a property value given a type and string value.

  The reverse of this method is just `str()` and `type()` of the Python value.

  Note that this does *not* support non-UTC offsets in ISO 8601-formatted
  datetime strings, e.g., the `-08:00` suffix in `2002-12-25 00:00:00-08:00`.
  It only supports `-00:00` and `+00:00` suffixes, which are UTC.

  Args:
    type_: A Python class.
    value_string: A string representation of the value of the property.

  Returns:
    An instance of `type`.

  Raises:
    `ValueError` if `type_` is datetime and `value_string` has a timezone
    offset.
  """
  if type_ == datetime.datetime:
    value_string = value_string.strip()

    if value_string[-6] in ('+', '-'):
      if value_string[-5:] == '00:00':
        value_string = value_string[:-6]
      else:

        raise ValueError('Non-UTC offsets in datetimes are not supported.')


    split = value_string.split('.')
    iso_date = split[0]
    microseconds = 0
    if len(split) > 1:
      microseconds = int(split[1])



    time_struct = time.strptime(iso_date, '%Y-%m-%d %H:%M:%S')[0:6]
    value = datetime.datetime(*(time_struct + (microseconds,)))
    return value
  elif type_ == Rating:

    return Rating(int(value_string))
  elif type_ == bool:
    return value_string == 'True'
  elif type_ == users.User:
    return users.User(value_string, _auth_domain)
  elif type_ == type(None):
    return None
  elif type_ in (Blob, EmbeddedEntity, ByteString):
    return type_(value_string.encode('utf-8'))
  return type_(value_string)


def ReferenceToKeyValue(key, id_resolver=None):
  """Converts a key into a comparable hashable `key` value.

  Args:
    key: The `entity_pb2.Reference` or `googledatastore.Key` from which to
        construct the key value.
    id_resolver: An optional `datastore_pbs.IdResolver`. Only necessary for
        `googledatastore.Key` values.
  Returns:
    A comparable and hashable representation of the given key that is
    compatible with one derived from a key property value.
  """
  if (datastore_pbs._CLOUD_DATASTORE_ENABLED
      and isinstance(key, googledatastore.Key)):
    v1_key = key
    key = entity_pb2.Reference()
    datastore_pbs.get_entity_converter(id_resolver).v1_to_v3_reference(v1_key,
                                                                       key)
  elif isinstance(key, entity_v4_pb2.Key):
    v4_key = key
    key = entity_pb2.Reference()
    datastore_pbs.get_entity_converter().v4_to_v3_reference(v4_key, key)

  if isinstance(key, entity_pb2.Reference):
    element_list = key.path.element
  elif isinstance(key, entity_pb2.PropertyValue.ReferenceValue):
    element_list = key.pathelement
  else:
    raise datastore_errors.BadArgumentError(
        'key arg expected to be entity_pb2.Reference or googledatastore.Key (%r)'
        % (key,))

  result = [
      entity_pb2.PropertyValue.REFERENCEVALUE_FIELD_NUMBER, key.app,
      key.name_space
  ]
  for element in element_list:
    result.append(element.type)
    if element.HasField('name'):
      result.append(element.name)
    else:
      result.append(element.id)
  return tuple(result)




def _isFloatNegative(value, encoded):
  if value == 0:
    return encoded[0] == 128
  return value < 0



def _encodeDoubleSortably(value):
  """Encode a double into a sortable byte buffer."""

  encoded = array.array('B')
  encoded.fromstring(struct.pack('>d', value))
  if _isFloatNegative(value, encoded):


    encoded[0] ^= 0xFF
    encoded[1] ^= 0xFF
    encoded[2] ^= 0xFF
    encoded[3] ^= 0xFF
    encoded[4] ^= 0xFF
    encoded[5] ^= 0xFF
    encoded[6] ^= 0xFF
    encoded[7] ^= 0xFF
  else:

    encoded[0] ^= 0x80
  return encoded


def PropertyValueToKeyValue(prop_value):
  """Converts an `entity_pb2.PropertyValue` into a comparable hashable `key` value.

  The values produces by this function mimic the native ordering of the
  datastore and uniquely identify the given `PropertyValue`.

  Args:
    prop_value: The `entity_pb2.PropertyValue` from which to construct the key
      value.

  Returns:
    A comparable and hashable representation of the given property value.
  """
  if not isinstance(prop_value, entity_pb2.PropertyValue):
    raise datastore_errors.BadArgumentError(
        'prop_value arg expected to be entity_pb2.PropertyValue (%r)' %
        (prop_value,))



  if prop_value.HasField('stringValue'):
    return (entity_pb2.PropertyValue.STRINGVALUE_FIELD_NUMBER,
            prop_value.stringValue)
  if prop_value.HasField('int64Value'):
    return (entity_pb2.PropertyValue.INT64VALUE_FIELD_NUMBER,
            prop_value.int64Value)
  if prop_value.HasField('booleanValue'):
    return (entity_pb2.PropertyValue.BOOLEANVALUE_FIELD_NUMBER,
            prop_value.booleanValue)
  if prop_value.HasField('doubleValue'):

    return (entity_pb2.PropertyValue.DOUBLEVALUE_FIELD_NUMBER,
            tuple(sortable_pb_encoder.EncodeDouble(prop_value.doubleValue)))
  if prop_value.HasField('pointvalue'):
    return (entity_pb2.PropertyValue.POINTVALUE_FIELD_NUMBER,
            prop_value.pointvalue.x, prop_value.pointvalue.y)
  if prop_value.HasField('referencevalue'):
    return ReferenceToKeyValue(prop_value.referencevalue)
  if prop_value.HasField('uservalue'):
    result = []
    uservalue = prop_value.uservalue
    if uservalue.HasField('email'):
      result.append((entity_pb2.PropertyValue.UserValue.EMAIL_FIELD_NUMBER,
                     uservalue.email))
    if uservalue.HasField('auth_domain'):
      result.append(
          (entity_pb2.PropertyValue.UserValue.AUTH_DOMAIN_FIELD_NUMBER,
           uservalue.auth_domain))
    if uservalue.HasField('nickname'):
      result.append((entity_pb2.PropertyValue.UserValue.NICKNAME_FIELD_NUMBER,
                     uservalue.nickname))
    if uservalue.HasField('gaiaid'):
      result.append((entity_pb2.PropertyValue.UserValue.GAIAID_FIELD_NUMBER,
                     uservalue.gaiaid))
    if uservalue.HasField('obfuscated_gaiaid'):
      result.append(
          (entity_pb2.PropertyValue.UserValue.OBFUSCATED_GAIAID_FIELD_NUMBER,
           uservalue.obfuscated_gaiaid))
    if uservalue.HasField('federated_identity'):
      result.append(
          (entity_pb2.PropertyValue.UserValue.FEDERATED_IDENTITY_FIELD_NUMBER,
           uservalue.federated_identity))
    if uservalue.HasField('federated_provider'):
      result.append(
          (entity_pb2.PropertyValue.UserValue.FEDERATED_PROVIDER_FIELD_NUMBER,
           uservalue.federated_provider))
    result.sort()
    return (entity_pb2.PropertyValue.USERVALUE_FIELD_NUMBER, tuple(result))
  return ()


def GetPropertyValueTag(value_pb):
  """Returns the tag constant associated with the given `entity_pb2.PropertyValue`."""
  if value_pb.HasField('booleanValue'):
    return entity_pb2.PropertyValue.BOOLEANVALUE_FIELD_NUMBER
  elif value_pb.HasField('doubleValue'):
    return entity_pb2.PropertyValue.DOUBLEVALUE_FIELD_NUMBER
  elif value_pb.HasField('int64Value'):
    return entity_pb2.PropertyValue.INT64VALUE_FIELD_NUMBER
  elif value_pb.HasField('pointvalue'):
    return entity_pb2.PropertyValue.POINTVALUE_FIELD_NUMBER
  elif value_pb.HasField('referencevalue'):
    return entity_pb2.PropertyValue.REFERENCEVALUE_FIELD_NUMBER
  elif value_pb.HasField('stringValue'):
    return entity_pb2.PropertyValue.STRINGVALUE_FIELD_NUMBER
  elif value_pb.HasField('uservalue'):
    return entity_pb2.PropertyValue.USERVALUE_FIELD_NUMBER
  else:
    return 0



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Primitives for dealing with datastore indexes.

Example `index.yaml` file:

```
indexes:

- kind: Cat
  ancestor: no
  properties:
  - name: name
  - name: age
    direction: desc

- kind: Cat
  properties:
  - name: name
    direction: ascending
  - name: whiskers
    direction: descending

- kind: Store
  ancestor: yes
  properties:
  - name: business
    direction: asc
  - name: owner
    direction: asc

- kind: Mountain
  properties:
  - name: name
  - name: location
    mode: geospatial
```

To learn more about the `index.yaml` file, see the configuration guide:
https://cloud.google.com/appengine/docs/standard/python3/configuring-datastore-indexes-with-index-yaml

"""








import copy
import itertools

import google

from ruamel import yaml
import six
from six.moves import zip

from google.appengine.api import appinfo
from google.appengine.api import datastore_types
from google.appengine.api import validation
from google.appengine.api import yaml_errors
from google.appengine.api import yaml_object
from google.appengine.datastore import datastore_pb
from google.appengine.datastore import entity_bytes_pb2 as entity_pb2








class Property(validation.Validated):
  """Representation for a property of an index as it appears in YAML.

  Attributes (all in string form):

  - `name`: Name of attribute to sort by.
  - `direction`: Direction of sort.
  - `mode`: How the property is indexed. Either `geospatial` or `None`
     (unspecified).
  """

  ATTRIBUTES = {
      'name': validation.Type(str, convert=False),
      'direction': validation.Optional([('asc', ('ascending',)),
                                        ('desc', ('descending',))]),
      'mode': validation.Optional(['geospatial'])
  }

  def IsAscending(self):






    return self.direction != 'desc'

  def CheckInitialized(self):
    if self.direction is not None and self.mode is not None:
      raise validation.ValidationError(
          'direction and mode are mutually exclusive')
    super(Property, self).CheckInitialized()


def PropertyPresenter(dumper, prop):
  """A PyYaml presenter for Property.

  It differs from the default by not outputting `mode: null` and direction when
  mode is specified. This is done in order to ensure backwards compatibility.

  Args:
    dumper: The Dumper object provided by PyYaml.
    prop: The Property object to serialize.

  Returns:
    A PyYaml object mapping.
  """


  prop_copy = copy.copy(prop)


  if prop.mode is None:
    del prop_copy.mode

  if prop.direction is None:
    del prop_copy.direction

  return dumper.represent_object(prop_copy)


class Index(validation.Validated):
  """Individual index definition.

  Order of the properties determines a given index's sort priority.

  Attributes:
    kind: Datastore kind that index belongs to.
    ancestors: Include ancestors in index.
    properties: Properties to be included.
  """

  ATTRIBUTES = {
      'kind': validation.Type(str, convert=False),
      'ancestor': validation.Type(bool, convert=False, default=False),
      'properties': validation.Optional(validation.Repeated(Property)),
  }

  def CheckInitialized(self):
    self._Normalize()
    super(Index, self).CheckInitialized()

  def _Normalize(self):
    if self.properties is None:
      return
    is_geo = any(x.mode == 'geospatial' for x in self.properties)
    for prop in self.properties:
      if is_geo:
        if prop.direction is not None:
          raise validation.ValidationError(
              'direction not supported in a geospatial index')
      else:

        if prop.IsAscending():
          prop.direction = 'asc'


class IndexDefinitions(validation.Validated):
  """Top level for index definition file.

  Attributes:
    indexes: List of Index definitions.
  """

  ATTRIBUTES = {
      appinfo.APPLICATION: validation.Optional(appinfo.APPLICATION_RE_STRING),
      'indexes': validation.Optional(validation.Repeated(Index)),
  }

index_yaml = yaml.YAML(typ='unsafe')
index_yaml.representer.add_representer(Property, PropertyPresenter)

if six.PY2:




  def _Py2UnicodeRepresenter(self, s):
    return self.represent_str(s.encode('utf-8'))
  index_yaml.representer.add_representer(unicode, _Py2UnicodeRepresenter)


def ParseIndexDefinitions(document, open_fn=None):
  """Parses an individual index definitions document from string or stream.

  Args:
    document: YAML document as a string or file-like stream.
    open_fn: Function for opening files. Unused.

  Raises:
    `EmptyConfigurationFile` when the configuration file is empty.
    `MultipleConfigurationFile` when the configuration file contains more than
    one document.

  Returns:
    Single parsed YAML file if one is defined, else `None`.
  """
  del open_fn
  try:
    return yaml_object.BuildSingleObject(IndexDefinitions, document)
  except yaml_errors.EmptyConfigurationFile:
    return None


def ParseMultipleIndexDefinitions(document):
  """Parses multiple index definitions documents from a string or stream.

  Args:
    document: YAML document as a string or file-like stream.

  Returns:
    A list of `datastore_index.IndexDefinitions` objects, one for each document.
  """
  return yaml_object.BuildObjects(IndexDefinitions, document)


def IndexDefinitionsToKeys(indexes):
  """Convert `IndexDefinitions` to a set of keys.

  Args:
    indexes: A `datastore_index.IndexDefinitions` instance, or `None`.

  Returns:
    A set of keys constructed from the argument, each key being a
    tuple of the form `(kind, ancestor, properties)` where `properties` is
    a tuple of `PropertySpec` objects.
  """
  keyset = set()
  if indexes is not None:
    if indexes.indexes:
      for index in indexes.indexes:
        keyset.add(IndexToKey(index))
  return keyset





def IndexToKey(index):
  """Converts Index to key.

  Args:
    index: A `datastore_index.Index` instance (not None).

  Returns:
    A tuple of the form `(kind, ancestor, properties)` where properties
    is a sequence of `PropertySpec` objects derived from the Index.
  """


  props = []
  if index.properties is not None:
    for prop in index.properties:
      props.append(PropertySpec(name=prop.name,
                                direction = (ASCENDING if prop.IsAscending()
                                             else DESCENDING)))
  return index.kind, index.ancestor, tuple(props)





class PropertySpec(object):
  """Index property attributes required to satisfy a query."""

  def __init__(self, name, direction=None, mode=None):
    assert direction is None or mode is None
    self._name = name
    self._direction = direction
    self._mode = mode

  @property
  def name(self):
    return self._name

  @property
  def direction(self):
    return self._direction

  @property
  def mode(self):
    return self._mode

  def __eq__(self, other):
    if not isinstance(other, PropertySpec):
      return NotImplemented
    return self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not self == other

  def __tuple(self):
    """Produces a tuple for comparison purposes."""
    direction = self._direction or entity_pb2.Index.Property.Direction.DIRECTION_UNSPECIFIED
    mode = self._mode or entity_pb2.Index.Property.Mode.MODE_UNSPECIFIED
    return (self._name, direction, mode)

  def __lt__(self, other):
    if not isinstance(other, PropertySpec):
      return NotImplemented
    return self.__tuple() < other.__tuple()

  def __le__(self, other):
    if not isinstance(other, PropertySpec):
      return NotImplemented
    return self.__tuple() <= other.__tuple()

  def __gt__(self, other):
    if not isinstance(other, PropertySpec):
      return NotImplemented
    return self.__tuple() > other.__tuple()

  def __ge__(self, other):
    if not isinstance(other, PropertySpec):
      return NotImplemented
    return self.__tuple() >= other.__tuple()

  def __hash__(self):
    return hash(('PropertySpec', self._name, self._direction, self._mode))

  def __repr__(self):
    builder = ['PropertySpec(name=%s' % self._name]
    if self._direction is not None:
      builder.append('direction=%s' %
                     entity_pb2.Index.Property.Direction.Name(self._direction))
    if self._mode is not None:
      builder.append('mode=%s' %
                     entity_pb2.Index.Property.Mode.Name(self._mode))
    return '%s)' % (', '.join(builder),)

  def Satisfies(self, other):
    """Determines whether existing index can satisfy requirements of a new query.

    Used in finding matching postfix with traditional "ordered" index specs.
    """
    assert isinstance(other, PropertySpec)
    if self._name != other._name:
      return False
    if self._mode is not None or other._mode is not None:



      return False
    if (other._direction is None):


      return True
    return self._direction == other._direction

  def CopyToIndexPb(self, pb):
    pb.name = self._name









    if (self._mode is None):
      pb.direction = self._direction or ASCENDING
    else:
      pb.mode = self._mode



GEOSPATIAL = entity_pb2.Index.Property.GEOSPATIAL
ASCENDING = entity_pb2.Index.Property.ASCENDING
DESCENDING = entity_pb2.Index.Property.DESCENDING


assert entity_pb2.Index.Property.ASCENDING == datastore_pb.Query.Order.ASCENDING
assert (
    entity_pb2.Index.Property.DESCENDING == datastore_pb.Query.Order.DESCENDING)


EQUALITY_OPERATORS = set([datastore_pb.Query.Filter.EQUAL])
INEQUALITY_OPERATORS = set([
    datastore_pb.Query.Filter.LESS_THAN,
    datastore_pb.Query.Filter.LESS_THAN_OR_EQUAL,
    datastore_pb.Query.Filter.GREATER_THAN,
    datastore_pb.Query.Filter.GREATER_THAN_OR_EQUAL
])
EXISTS_OPERATORS = set([datastore_pb.Query.Filter.EXISTS])


def Normalize(filters, orders, exists):
  """Normalizes filter and order query components.

  The resulting components have the same effect as the given components if used
  in a query.

  Args:
    filters: The filters set on the query.
    orders: The orders set on the query.
    exists: The names of properties that require an exists filter if
      not already specified.

  Returns:
    `(filter, orders)` the reduced set of filters and orders.
  """


  eq_properties = set()
  inequality_properties = set()


  for f in filters:
    if f.op == datastore_pb.Query.Filter.IN and len(f.property) == 1:
      f.op = datastore_pb.Query.Filter.EQUAL
    if f.op in EQUALITY_OPERATORS:
      eq_properties.add(f.property[0].name)
    elif f.op in INEQUALITY_OPERATORS:
      inequality_properties.add(f.property[0].name)

  eq_properties -= inequality_properties


  remove_set = eq_properties.copy()
  new_orders = []
  for o in orders:
    if o.property not in remove_set:
      remove_set.add(o.property)
      new_orders.append(o)
  orders = new_orders

  remove_set.update(inequality_properties)


  new_filters = []
  for f in filters:
    if f.op not in EXISTS_OPERATORS:
      new_filters.append(f)
      continue
    name = f.property[0].name
    if name not in remove_set:
      remove_set.add(name)
      new_filters.append(f)


  for prop in exists:
    if prop not in remove_set:
      remove_set.add(prop)
      new_filter = datastore_pb.Query.Filter()
      new_filter.op = datastore_pb.Query.Filter.EXISTS
      new_prop = new_filter.property.add()
      new_prop.name = prop
      new_prop.multiple = False
      new_prop.ClearField('value')
      new_filters.append(new_filter)

  filters = new_filters




  if datastore_types.KEY_SPECIAL_PROPERTY in eq_properties:
    orders = []



  new_orders = []
  for o in orders:
    if o.property == datastore_types.KEY_SPECIAL_PROPERTY:
      new_orders.append(o)
      break
    new_orders.append(o)
  orders = new_orders

  return (filters, orders)


def RemoveNativelySupportedComponents(filters, orders, exists):
  """Removes query components that are natively supported by the datastore.

  The resulting filters and orders should not be used in an actual query.

  Args:
    filters: The filters set on the query.
    orders: The orders set on the query.
    exists: The names of properties that require an exists filter if
      not already specified.

  Returns:
    `(filters, orders)` the reduced set of filters and orders
  """
  (filters, orders) = Normalize(filters, orders, exists)

  for f in filters:
    if f.op in EXISTS_OPERATORS:



      return (filters, orders)


  has_key_desc_order = False
  if orders and orders[-1].property == datastore_types.KEY_SPECIAL_PROPERTY:
    if orders[-1].direction == ASCENDING:
      orders = orders[:-1]
    else:
      has_key_desc_order = True






  if not has_key_desc_order:
    for f in filters:
      if (f.op in INEQUALITY_OPERATORS and
          f.property[0].name != datastore_types.KEY_SPECIAL_PROPERTY):
        break
    else:
      filters = [
          f for f in filters
          if f.property[0].name != datastore_types.KEY_SPECIAL_PROPERTY
      ]

  return (filters, orders)


def CompositeIndexForQuery(query):
  """Returns the composite index needed for a query.

  A query is translated into a tuple, as follows:

  - The first item is the kind string, or `None` if we're not filtering
    on kind (see below).

  - The second item is a bool giving whether the query specifies an
    ancestor.

  - After that come `(property, ASCENDING)` pairs for those Filter
    entries whose operator is `EQUAL` or `IN`. Since the order of these
    doesn't matter, they are sorted by property name to normalize them
    in order to avoid duplicates.

  - After that comes at most one `(property, ASCENDING)` pair for a
    Filter entry whose operator is on of the four inequalities. There
    can be at most one of these.

  - After that come all the `(property, direction)` pairs for the Order
    entries, in the order given in the query. Exceptions:

    - (a) If there is a Filter entry with an inequality operator that matches
          the first Order entry, the first order pair is omitted (or,
          equivalently, in this case the inequality pair is omitted).
    - (b) If an Order entry corresponds to an equality filter, it is ignored
          (since there will only ever be one value returned).
    - (c) If there is an equality filter on `__key__` all orders are dropped
          (since there will be at most one result returned).
    - (d) If there is an order on `__key__` all further orders are dropped
          (since keys are unique).
    - (e) Orders on `__key__` `ASCENDING` are dropped (since this is supported
          natively by the datastore).

  - Finally, if there are Filter entries whose operator is `EXISTS`, and
    whose property names are not already listed, they are added, with
    the direction set to `ASCENDING`.

  This algorithm should consume all Filter and Order entries.

  Additional notes:

  - The low-level implementation allows queries that don't specify a
    kind; but the Python `API` doesn't support this yet.

  - If there's an inequality filter and one or more sort orders, the
    first sort order *must* match the inequality filter.

  - The following indexes are always built in and should be suppressed:

    - Query on kind only;
    - Query on kind and one filter *or* one order;
    - Query on ancestor only, without kind (not exposed in Python yet);
    - Query on kind and equality filters only, no order (with or without
      ancestor).

  - While the protocol buffer allows a Filter to contain multiple
    properties, we don't use this.  It is only needed for the `IN` operator
    but this is (currently) handled on the client side, so in practice
    each Filter is expected to have exactly one property.

  Args:
    query: A `datastore_pb.Query` instance.

  Returns:
    A tuple of the form `(required, kind, ancestor, properties)`.
      `required`: Boolean, whether the index is required;
      `kind`: The kind or None;
      `ancestor`: True if this is an ancestor query;
      `properties`: A tuple consisting of:
      - The prefix, represented by a set of property names.
      - The postfix, represented by a tuple consisting of any number of:
        - Sets of property names or `PropertySpec` objects: these
          properties can appear in any order.
        - Sequences of `PropertySpec` objects: Indicates the properties
          must appear in the given order, with the specified direction (if
          specified in the `PropertySpec`).
  """
  required = True


  kind = query.kind
  ancestor = query.HasField('ancestor')
  filters = query.filter
  orders = query.order



  for filter in filters:
    assert filter.op != datastore_pb.Query.Filter.IN, 'Filter.op==IN'
    nprops = len(filter.property)
    assert nprops == 1, 'Filter has %s properties, expected 1' % nprops
    if filter.op == datastore_pb.Query.Filter.CONTAINED_IN_REGION:
      return CompositeIndexForGeoQuery(query)

  if not kind:


    required = False

  exists = list(query.property_name)
  exists.extend(query.group_by_property_name)

  filters, orders = RemoveNativelySupportedComponents(filters, orders, exists)


  eq_filters = [f for f in filters if f.op in EQUALITY_OPERATORS]
  ineq_filters = [f for f in filters if f.op in INEQUALITY_OPERATORS]
  exists_filters = [f for f in filters if f.op in EXISTS_OPERATORS]
  assert (len(eq_filters) + len(ineq_filters) +
          len(exists_filters)) == len(filters), 'Not all filters used'

  if (kind and not ineq_filters and not exists_filters and
      not orders):



    names = set(f.property[0].name for f in eq_filters)
    if not names.intersection(datastore_types._SPECIAL_PROPERTIES):
      required = False



  ineq_property = None
  if ineq_filters:
    for filter in ineq_filters:
      if (filter.property[0].name ==
          datastore_types._UNAPPLIED_LOG_TIMESTAMP_SPECIAL_PROPERTY):
        continue
      if not ineq_property:
        ineq_property = filter.property[0].name
      else:
        assert filter.property[0].name == ineq_property



  group_by_props = set(query.group_by_property_name)


  prefix = frozenset(f.property[0].name for f in eq_filters)

  postfix_ordered = [
      PropertySpec(name=order.property, direction=order.direction)
      for order in orders
  ]


  postfix_group_by = frozenset(f.property[0].name
                               for f in exists_filters
                               if f.property[0].name in group_by_props)

  postfix_unordered = frozenset(f.property[0].name
                                for f in exists_filters
                                if f.property[0].name not in group_by_props)


  if ineq_property:
    if orders:


      assert ineq_property == orders[0].property
    else:
      postfix_ordered.append(PropertySpec(name=ineq_property))

  property_count = (len(prefix) + len(postfix_ordered) + len(postfix_group_by)
                    + len(postfix_unordered))
  if kind and not ancestor and property_count <= 1:


    required = False


    if postfix_ordered:
      prop = postfix_ordered[0]
      if (prop.name == datastore_types.KEY_SPECIAL_PROPERTY and
          prop.direction == DESCENDING):
        required = True


  props = prefix, (tuple(postfix_ordered), postfix_group_by, postfix_unordered)
  return required, kind, ancestor, props


def CompositeIndexForGeoQuery(query):
  """Builds a descriptor for a composite index needed for a geo query.

  Args:
    query: A `datastore_pb.Query` instance.

  Returns:
    A tuple in the same form as produced by `CompositeIndexForQuery`.
  """
  required = True
  kind = query.kind

  assert not query.HasField('ancestor')
  ancestor = False
  preintersection_props = set()
  geo_props = set()
  for filter_ in query.filter:
    name = filter_.property[0].name
    if filter_.op == datastore_pb.Query.Filter.EQUAL:
      preintersection_props.add(PropertySpec(name=name))
    else:

      assert filter_.op == datastore_pb.Query.Filter.CONTAINED_IN_REGION
      geo_props.add(PropertySpec(name=name, mode=GEOSPATIAL))

  prefix = frozenset(preintersection_props)
  postfix = (frozenset(geo_props),)
  return required, kind, ancestor, (prefix, postfix)


def GetRecommendedIndexProperties(properties):
  """Converts the properties returned by `datastore_index.CompositeIndexForQuery` into a recommended list of index properties with the desired constraints.

  Sets (of property names or `PropertySpec` objects) are sorted, so as to
  normalize them.

  Args:
    properties: See `datastore_index.CompositeIndexForQuery`.

  Returns:
    A tuple of `PropertySpec` objects.

  """

  prefix, postfix = properties
  result = []
  for sub_list in itertools.chain((prefix,), postfix):
    if isinstance(sub_list, (frozenset, set)):



      result.extend([(p if isinstance(p, PropertySpec)
                      else PropertySpec(name=p)) for p in sorted(sub_list)])
    else:
      result.extend([(PropertySpec(name=p.name, direction=ASCENDING)
                      if p.direction is None else p) for p in sub_list])

  return tuple(result)


def _MatchPostfix(postfix_props, index_props):
  """Matches a postfix constraint with an existing index.

  postfix_props constraints are specified through a list of:
  - sets of string: any order any direction;
  - list of tuples(string, direction): the given order, and, if specified, the
  given direction.

  For example (PropertySpec objects shown here in their legacy shorthand form):
    [set('A', 'B'), [('C', None), ('D', ASC)]]
  matches:
    [('F', ASC), ('B', ASC), ('A', DESC), ('C', DESC), ('D', ASC)]
  with a return value of [('F', ASC)], but does not match:
    [('F', ASC), ('A', DESC), ('C', DESC), ('D', ASC)]
    [('B', ASC), ('F', ASC), ('A', DESC), ('C', DESC), ('D', ASC)]
    [('F', ASC), ('B', ASC), ('A', DESC), ('C', DESC), ('D', DESC)]

  Args:
    postfix_props: A tuple of sets and lists, as output by
        `CompositeIndexForQuery`. They should define the requirements for the
        postfix of the index.
    index_props: A list of `PropertySpec` objects that
        define the index to try and match.

  Returns:
    The list of `PropertySpec` objects that define the prefix properties
    in the given index.  None if the constraints could not be
    satisfied.

  """


  index_props_rev = reversed(index_props)
  for property_group in reversed(postfix_props):
    index_group_iter = itertools.islice(index_props_rev, len(property_group))
    if isinstance(property_group, (frozenset, set)):

      index_group = set(prop.name for prop in index_group_iter)
      if index_group != property_group:
        return None
    else:


      index_group = list(index_group_iter)
      if len(index_group) != len(property_group):
        return None
      for candidate, spec in zip(index_group, reversed(property_group)):
        if not candidate.Satisfies(spec):
          return None
  remaining = list(index_props_rev)
  remaining.reverse()
  return remaining


def MinimalCompositeIndexForQuery(query, index_defs):
  """Computes the minimal composite index for this query.

  Unlike `datastore_index.CompositeIndexForQuery`, this function takes into
  account indexes that already exist in the system.

  Args:
    query: The `datastore_pb.Query` to compute suggestions for.
    index_defs: A list of `datastore_index.Index` objects that already exist.

  Returns:
    `None` if no index is needed, otherwise the minimal index in the form
    `(is_most_efficient, kind, ancestor, properties)`. Where `is_most_efficient`
    is a boolean denoting if the suggested index is the most efficient (i.e.,
    the one returned by `datastore_index.CompositeIndexForQuery`). kind and
    ancestor are the same variables returned by
    `datastore_index.CompositeIndexForQuery`. properties is a tuple consisting
    of the prefix and postfix properties returend by
    `datastore_index.CompositeIndexForQuery`.
  """

  required, kind, ancestor, (prefix, postfix) = CompositeIndexForQuery(query)

  if not required:
    return None


  remaining_dict = {}

  for definition in index_defs:
    if (kind != definition.kind or

        (not ancestor and definition.ancestor)):
      continue

    _, _, index_props = IndexToKey(definition)

    index_prefix = _MatchPostfix(postfix, index_props)

    if index_prefix is None:

      continue

    remaining_index_props = set([prop.name for prop in index_prefix])

    if remaining_index_props - prefix:

      continue


    index_postfix = tuple(index_props[len(index_prefix):])
    remaining = remaining_dict.get(index_postfix)
    if remaining is None:
      remaining = prefix.copy(), ancestor


    props_remaining, ancestor_remaining = remaining
    props_remaining -= remaining_index_props
    if definition.ancestor:
      ancestor_remaining = False

    if not (props_remaining or ancestor_remaining):
      return None

    if (props_remaining, ancestor_remaining) == remaining:
      continue


    remaining_dict[index_postfix] = (props_remaining, ancestor_remaining)

  if not remaining_dict:
    return (True, kind, ancestor, (prefix, postfix))

  def calc_cost(minimal_props, minimal_ancestor):
    result = len(minimal_props)
    if minimal_ancestor:
      result += 2
    return result


  minimal_postfix, remaining = remaining_dict.popitem()
  minimal_props, minimal_ancestor = remaining
  minimal_cost = calc_cost(minimal_props, minimal_ancestor)
  for index_postfix, (props_remaining,
                      ancestor_remaining) in (list(remaining_dict.items())):
    cost = calc_cost(props_remaining, ancestor_remaining)
    if cost < minimal_cost:
      minimal_cost = cost
      minimal_postfix = index_postfix
      minimal_props = props_remaining
      minimal_ancestor = ancestor_remaining


  props = frozenset(minimal_props), (minimal_postfix, frozenset(), frozenset())
  return False, kind, minimal_ancestor, props






def IndexYamlForQuery(kind, ancestor, props):
  """Returns the composite index definition `YAML` needed for a query.

  Given a query, the arguments for this method can be computed with:

    ```
    _, kind, ancestor, props = datastore_index.CompositeIndexForQuery(query)
    props = datastore_index.GetRecommendedIndexProperties(props)
    ```

  Args:
    kind: The kind or `None`.
    ancestor: `True` if this is an ancestor query, `False` otherwise.
    props: `PropertySpec` objects.

  Returns:
    A string with the YAML for the composite index needed by the query.
  """

  serialized_yaml = []
  serialized_yaml.append('- kind: %s' % kind)
  if ancestor:
    serialized_yaml.append('  ancestor: yes')
  if props:
    serialized_yaml.append('  properties:')
    for prop in props:
      serialized_yaml.append('  - name: %s' % prop.name)
      if prop.direction == DESCENDING:
        serialized_yaml.append('    direction: desc')
      if prop.mode is GEOSPATIAL:
        serialized_yaml.append('    mode: geospatial')
  return '\n'.join(serialized_yaml)


def IndexXmlForQuery(kind, ancestor, props):
  """Returns the composite index definition XML needed for a query.

  Given a query, the arguments for this method can be computed with:

    ```
    _, kind, ancestor, props = datastore_index.CompositeIndexForQuery(query)
    props = datastore_index.GetRecommendedIndexProperties(props)
    ```

  Args:
    kind: The kind or `None`
    ancestor: `True` if this is an ancestor query, `False` otherwise.
    props: `PropertySpec` objects.

  Returns:
    A string with the XML for the composite index needed by the query.
  """

  serialized_xml = []


  is_geo = any(p.mode is GEOSPATIAL for p in props)
  if is_geo:
    ancestor_clause = ''
  else:
    ancestor_clause = 'ancestor="%s"' % ('true' if ancestor else 'false',)
  serialized_xml.append('  <datastore-index kind="%s" %s>'
                        % (kind, ancestor_clause))
  for prop in props:
    if prop.mode is GEOSPATIAL:
      qual = ' mode="geospatial"'
    elif is_geo:
      qual = ''
    else:

      qual = ' direction="%s"' % ('desc' if prop.direction == DESCENDING
                                  else 'asc')
    serialized_xml.append('    <property name="%s"%s />' % (prop.name, qual))
  serialized_xml.append('  </datastore-index>')
  return '\n'.join(serialized_xml)


def IndexDefinitionToProto(app_id, index_definition):
  """Transforms individual Index definition to protocol buffer.

  Args:
    app_id: Application id for new protocol buffer `CompositeIndex`.
    index_definition: `datastore_index.Index` object to transform.

  Returns:
    New `entity_pb2.CompositeIndex` with default values set and index
    information filled in.
  """
  proto = entity_pb2.CompositeIndex()

  proto.app_id = app_id
  proto.id = 0
  proto.state = entity_pb2.CompositeIndex.WRITE_ONLY

  definition_proto = proto.definition
  definition_proto.entity_type = index_definition.kind
  definition_proto.ancestor = index_definition.ancestor

  if index_definition.properties is not None:
    is_geo = any(x.mode == 'geospatial' for x in index_definition.properties)
    for prop in index_definition.properties:
      prop_proto = definition_proto.property.add()
      prop_proto.name = prop.name

      if prop.mode == 'geospatial':
        prop_proto.mode = entity_pb2.Index.Property.GEOSPATIAL
      elif is_geo:




        pass
      elif prop.IsAscending():
        prop_proto.direction = entity_pb2.Index.Property.ASCENDING
      else:
        prop_proto.direction = entity_pb2.Index.Property.DESCENDING

  return proto


def IndexDefinitionsToProtos(app_id, index_definitions):
  """Transforms multiple index definitions to composite index records.

  Args:
    app_id: Application id for new protocol buffer `CompositeIndex`.
    index_definition: A list of datastore_index.Index objects to transform.

  Returns:
    A list of transformed entity_pb2.CompositeIndex entities with default values
    set and index information filled in.
  """
  return [IndexDefinitionToProto(app_id, index)
          for index in index_definitions]


def ProtoToIndexDefinition(proto):
  """Transforms individual index protocol buffer to index definition.

  Args:
    proto: An instance of `entity_pb2.CompositeIndex` to transform.

  Returns:
    A new instance of `datastore_index.Index`.
  """
  properties = []
  proto_index = proto.definition
  for prop_proto in proto_index.property:
    prop_definition = Property(name=prop_proto.name)

    if prop_proto.mode == entity_pb2.Index.Property.GEOSPATIAL:
      prop_definition.mode = 'geospatial'
    elif prop_proto.direction == entity_pb2.Index.Property.DESCENDING:
      prop_definition.direction = 'desc'
    elif prop_proto.direction == entity_pb2.Index.Property.ASCENDING:
      prop_definition.direction = 'asc'

    properties.append(prop_definition)

  index = Index(kind=proto_index.entity_type, properties=properties)
  if proto_index.ancestor:
    index.ancestor = True
  return index


def ProtosToIndexDefinitions(protos):
  """Returns index definitions, which were transformed from multiple index protocol buffers.

  Args:
    A list of `entity_pb2.Index` records.
  """
  return [ProtoToIndexDefinition(definition) for definition in protos]



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#google/appengine/api/api_base.proto\x12\x15google.appengine.base\"\x1c\n\x0bStringProto\x12\r\n\x05value\x18\x01 \x02(\t\"\x1f\n\x0eInteger32Proto\x12\r\n\x05value\x18\x01 \x02(\x05\"\x1f\n\x0eInteger64Proto\x12\r\n\x05value\x18\x01 \x02(\x03\"\x1a\n\tBoolProto\x12\r\n\x05value\x18\x01 \x02(\x08\"\x1c\n\x0b\x44oubleProto\x12\r\n\x05value\x18\x01 \x02(\x01\"\x1f\n\nBytesProto\x12\x11\n\x05value\x18\x01 \x02(\x0c\x42\x02\x08\x01\"\x0b\n\tVoidProtoB,\n\x1f\x63om.google.google.appengine.apiB\tApiBasePb')



_STRINGPROTO = DESCRIPTOR.message_types_by_name['StringProto']
_INTEGER32PROTO = DESCRIPTOR.message_types_by_name['Integer32Proto']
_INTEGER64PROTO = DESCRIPTOR.message_types_by_name['Integer64Proto']
_BOOLPROTO = DESCRIPTOR.message_types_by_name['BoolProto']
_DOUBLEPROTO = DESCRIPTOR.message_types_by_name['DoubleProto']
_BYTESPROTO = DESCRIPTOR.message_types_by_name['BytesProto']
_VOIDPROTO = DESCRIPTOR.message_types_by_name['VoidProto']
StringProto = _reflection.GeneratedProtocolMessageType('StringProto', (_message.Message,), {
  'DESCRIPTOR' : _STRINGPROTO,
  '__module__' : 'google.appengine.api.api_base_pb2'

  })
_sym_db.RegisterMessage(StringProto)

Integer32Proto = _reflection.GeneratedProtocolMessageType('Integer32Proto', (_message.Message,), {
  'DESCRIPTOR' : _INTEGER32PROTO,
  '__module__' : 'google.appengine.api.api_base_pb2'

  })
_sym_db.RegisterMessage(Integer32Proto)

Integer64Proto = _reflection.GeneratedProtocolMessageType('Integer64Proto', (_message.Message,), {
  'DESCRIPTOR' : _INTEGER64PROTO,
  '__module__' : 'google.appengine.api.api_base_pb2'

  })
_sym_db.RegisterMessage(Integer64Proto)

BoolProto = _reflection.GeneratedProtocolMessageType('BoolProto', (_message.Message,), {
  'DESCRIPTOR' : _BOOLPROTO,
  '__module__' : 'google.appengine.api.api_base_pb2'

  })
_sym_db.RegisterMessage(BoolProto)

DoubleProto = _reflection.GeneratedProtocolMessageType('DoubleProto', (_message.Message,), {
  'DESCRIPTOR' : _DOUBLEPROTO,
  '__module__' : 'google.appengine.api.api_base_pb2'

  })
_sym_db.RegisterMessage(DoubleProto)

BytesProto = _reflection.GeneratedProtocolMessageType('BytesProto', (_message.Message,), {
  'DESCRIPTOR' : _BYTESPROTO,
  '__module__' : 'google.appengine.api.api_base_pb2'

  })
_sym_db.RegisterMessage(BytesProto)

VoidProto = _reflection.GeneratedProtocolMessageType('VoidProto', (_message.Message,), {
  'DESCRIPTOR' : _VOIDPROTO,
  '__module__' : 'google.appengine.api.api_base_pb2'

  })
_sym_db.RegisterMessage(VoidProto)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\037com.google.google.appengine.apiB\tApiBasePb'
  _BYTESPROTO.fields_by_name['value']._options = None
  _BYTESPROTO.fields_by_name['value']._serialized_options = b'\010\001'
  _STRINGPROTO._serialized_start=62
  _STRINGPROTO._serialized_end=90
  _INTEGER32PROTO._serialized_start=92
  _INTEGER32PROTO._serialized_end=123
  _INTEGER64PROTO._serialized_start=125
  _INTEGER64PROTO._serialized_end=156
  _BOOLPROTO._serialized_start=158
  _BOOLPROTO._serialized_end=184
  _DOUBLEPROTO._serialized_start=186
  _DOUBLEPROTO._serialized_end=214
  _BYTESPROTO._serialized_start=216
  _BYTESPROTO._serialized_end=247
  _VOIDPROTO._serialized_start=249
  _VOIDPROTO._serialized_end=260




#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Test utilities for API modules.

Classes:
  ModuleInterfaceTest: Test framework for developing public modules.
"""

import __future__
import types




class ModuleInterfaceTest(object):
  """Test to ensure module interface is carefully constructed.

  A module interface is the set of public objects listed in the module __all__
  attribute.  Modules that will be used by the public should have this interface
  carefully declared.  At all times, the __all__ attribute should have objects
  intended to be used by the public and other objects in the module should be
  considered unused.

  Protected attributes (those beginning with '_') and other imported modules
  should not be part of this set of variables.  An exception is for variables
  that begin and end with '__' which are implicitly part of the interface
  (eg. __name__, __file__, __all__ itself, etc.).

  Modules that are imported in to the tested modules are an exception and may
  be left out of the __all__ definition. The test is done by checking the value
  of what would otherwise be a public name and not allowing it to be exported
  if it is an instance of a module.  Modules that are explicitly exported are
  for the time being not permitted.

  To use this test class a module should define a new class that inherits first
  from ModuleInterfaceTest and then from absltest.TestCase.  No other tests
  should be added to this test case, making the order of inheritance less
  important, but if setUp for some reason is overidden, it is important that
  ModuleInterfaceTest is first in the list so that its setUp method is
  invoked.

  Multiple inheretance is required so that ModuleInterfaceTest is not itself
  a test, and is not itself executed as one.

  The test class is expected to have the following class attributes defined:

    MODULE: A reference to the module that is being validated for interface
      correctness.

  Example:
    Module definition (hello.py):

      import sys

      __all__ = ['hello']

      def _get_outputter():
        return sys.stdout

      def hello():
        _get_outputter().write('Hello\n')

    Test definition:

      from google.appengine.api \
          import module_testutil

      from absl import app
      from absl.testing import absltest

      import hello

      class ModuleInterfaceTest(module_testutil.ModuleInterfaceTest,
                                absltest.TestCase):

        MODULE = hello


      class HelloTest(absltest.TestCase):
        ... Test 'hello' module ...


      if __name__ == '__main__':
        absltest.main()
  """

  def setUp(self):
    """Set up makes sure that MODULE and IMPORTED_MODULES is defined.

    This is a basic configuration test for the test itself so does not
    get it's own test case.
    """
    if not hasattr(self, 'MODULE'):
      self.fail(
          "You must define 'MODULE' on ModuleInterfaceTest sub-class %s." %
          type(self).__name__)

  def testAllExist(self):
    """Test that all attributes defined in __all__ exist."""
    missing_attributes = []
    for attribute in self.MODULE.__all__:
      if not hasattr(self.MODULE, attribute):
        missing_attributes.append(attribute)
    if missing_attributes:
      self.fail('%s of __all__ are not defined in module.' %
                missing_attributes)

  def testAllExported(self):
    """Test that all public attributes not imported are in __all__."""
    missing_attributes = []
    for attribute in dir(self.MODULE):
      if not attribute.startswith('_'):
        if attribute not in self.MODULE.__all__:
          attribute_value = getattr(self.MODULE, attribute)
          if isinstance(attribute_value, types.ModuleType):
            continue
          if isinstance(attribute_value, __future__._Feature):
            continue
          missing_attributes.append(attribute)
    if missing_attributes:
      self.fail('%s are not modules and not defined in __all__.' %
                missing_attributes)

  def testNoExportedProtectedVariables(self):
    """Test that there are no protected variables listed in __all__."""
    protected_variables = []
    for attribute in self.MODULE.__all__:
      if attribute.startswith('_'):
        protected_variables.append(attribute)
    if protected_variables:
      self.fail('%s are protected variables and may not be exported.' %
                protected_variables)

  def testNoExportedModules(self):
    """Test that no modules exist in __all__."""
    exported_modules = []
    for attribute in self.MODULE.__all__:
      try:
        value = getattr(self.MODULE, attribute)
      except AttributeError:

        pass
      else:
        if isinstance(value, types.ModuleType):
          exported_modules.append(attribute)
    if exported_modules:
      self.fail('%s are modules and may not be exported.' % exported_modules)



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Builder for mapping YAML documents to object instances.

`ObjectBuilder` is responsible for mapping a YAML document to classes defined
using the validation mechanism (see `google.appengine.api.validation.py`).
"""







import google

from ruamel import yaml

from google.appengine.api import validation
from google.appengine.api import yaml_builder
from google.appengine.api import yaml_errors
from google.appengine.api import yaml_listener




class _ObjectMapper(object):
  """Wrapper used for mapping attributes from a yaml file to an object.

  This wrapper is required because objects do not know what property they are
  associated with a creation time, and therefore can not be instantiated
  with the correct class until they are mapped to their parents.
  """

  def __init__(self):
    """Object mapper starts off with empty value."""
    self.value = None
    self.seen = set()

  def set_value(self, value):
    """Set value of instance to map to.

    Args:
      value: Instance that this mapper maps to.
    """
    self.value = value

  def see(self, key):
    if key in self.seen:
      raise yaml_errors.DuplicateAttribute("Duplicate attribute '%s'." % key)
    self.seen.add(key)


class _ObjectSequencer(object):
  """Wrapper used for building sequences from a yaml file to a list.

  This wrapper is required because objects do not know what property they are
  associated with a creation time, and therefore can not be instantiated
  with the correct class until they are mapped to their parents.
  """

  def __init__(self):
    """Object sequencer starts off with empty value."""
    self.value = []
    self.constructor = None

  def set_constructor(self, constructor):
    """Set object used for constructing new sequence instances.

    Args:
      constructor: Callable which can accept no arguments.  Must return
        an instance of the appropriate class for the container.
    """
    self.constructor = constructor


class ObjectBuilder(yaml_builder.Builder):
  """Builder used for constructing validated objects.

  Given a class that implements `validation.ValidatedBase`, it will parse a YAML
  document and attempt to build an instance of the class.
  `ObjectBuilder` will only map YAML fields that are accepted by the
  `ValidatedBase`'s `GetValidator` function. Lists are mapped to validated.
  Repeated attributes and maps are mapped to `validated.Type` properties.

  For a YAML map to be compatible with a class, the class must have a
  constructor that can be called with no parameters. If the provided type
  does not have such a constructor a parse time error will occur.

  The constructor initializes the validated object builder.
  """

  def __init__(self, default_class):
    """Constructor.

    Args:
      default_class: Class that is instantiated upon the detection of a new
        document. An instance of this class will act as the document itself.
    """
    self.default_class = default_class

  def _GetRepeated(self, attribute):
    """Get the ultimate type of a repeated validator.

    Looks for an instance of validation.Repeated, returning its constructor.

    Args:
      attribute: Repeated validator attribute to find type for.

    Returns:
      The expected class of of the Type validator, otherwise object.
    """
    if isinstance(attribute, validation.Optional):
      attribute = attribute.validator
    if isinstance(attribute, validation.Repeated):
      return attribute.constructor
    return object

  def BuildDocument(self):
    """Instantiate new root validated object.

    Returns:
      New instance of validated object.
    """
    return self.default_class()

  def BuildMapping(self, top_value):
    """New instance of object mapper for opening map scope.

    Args:
      top_value: Parent of nested object.

    Returns:
      New instance of object mapper.
    """
    result = _ObjectMapper()


    if isinstance(top_value, self.default_class):
      result.value = top_value
    return result

  def EndMapping(self, top_value, mapping):
    """When leaving scope, ensures the new object is initialized.

    This method is mainly for picking up on any missing required attributes.

    Args:
      top_value: Parent of closing mapping object.
      mapping: `_ObjectMapper` instance that is leaving scope.
    """


    if not hasattr(mapping.value, 'CheckInitialized'):
      raise validation.ValidationError('Cannot convert map to non-map value.')

    try:
      mapping.value.CheckInitialized()
    except validation.ValidationError:

      raise
    except Exception as e:




      try:
        error_str = str(e)
      except Exception:
        error_str = '<unknown>'


      raise validation.ValidationError(error_str, e)

  def BuildSequence(self, top_value):
    """New instance of object sequence.

    Args:
      top_value: Object that contains the new sequence.

    Returns:
      A new `_ObjectSequencer` instance.
    """
    return _ObjectSequencer()

  def MapTo(self, subject, key, value):
    """Maps key-value pair to an objects attribute.

    Args:
      subject: `_ObjectMapper` of object that will receive new attribute.
      key: Key of attribute.
      value: Value of new attribute.

    Raises:
      `UnexpectedAttribute` when the key is not a validated attribute of
      the subject value class.
    """
    assert isinstance(subject.value, validation.ValidatedBase)

    try:
      attribute = subject.value.GetValidator(key)
    except validation.ValidationError as err:
      raise yaml_errors.UnexpectedAttribute(err)

    if isinstance(value, _ObjectMapper):


      value.set_value(attribute.expected_type())
      value = value.value
    elif isinstance(value, _ObjectSequencer):

      value.set_constructor(self._GetRepeated(attribute))
      value = value.value

    subject.see(key)
    try:
      subject.value.Set(key, value)
    except validation.ValidationError as e:




      try:
        error_str = str(e)
      except Exception:
        error_str = '<unknown>'

      try:
        value_str = str(value)
      except Exception:
        value_str = '<unknown>'


      e.message = ("Unable to assign value '%s' to attribute '%s':\n%s" %
                   (value_str, key, error_str))
      raise e
    except Exception as e:
      try:
        error_str = str(e)
      except Exception:
        error_str = '<unknown>'

      try:
        value_str = str(value)
      except Exception:
        value_str = '<unknown>'


      message = ("Unable to assign value '%s' to attribute '%s':\n%s" %
                 (value_str, key, error_str))
      raise validation.ValidationError(message, e)

  def AppendTo(self, subject, value):
    """Append a value to a sequence.

    Args:
      subject: `_ObjectSequence` that is receiving new value.
      value: Value that is being appended to sequence.
    """
    if isinstance(value, _ObjectMapper):

      value.set_value(subject.constructor())
      subject.value.append(value.value)
    else:

      subject.value.append(value)


def BuildObjects(default_class, stream, loader=yaml.loader.SafeLoader):
  """Build objects from stream.

  Handles the basic case of loading all the objects from a stream.

  Args:
    default_class: Class that is instantiated upon the detection of a new
      document.  An instance of this class will act as the document itself.
    stream: String document or open file object to process as per the
      `yaml.parse` method.  Any object that implements a `read()` method which
      returns a string document will work with the YAML parser.
    loader_class: Used for dependency injection.

  Returns:
    List of `default_class` instances parsed from the stream.
  """
  builder = ObjectBuilder(default_class)
  handler = yaml_builder.BuilderHandler(builder)
  listener = yaml_listener.EventListener(handler)

  listener.Parse(stream, loader, version=(1, 1))
  return handler.GetResults()


def BuildSingleObject(default_class, stream, loader=yaml.loader.SafeLoader):
  """Build object from stream.

  Handles the basic case of loading a single object from a stream.

  Args:
    default_class: Class that is instantiated upon the detection of a new
      document.  An instance of this class will act as the document itself.
    stream: String document or open file object to process as per the
      yaml.parse method.  Any object that implements a 'read()' method which
      returns a string document will work with the YAML parser.
    loader_class: Used for dependency injection.
  """
  definitions = BuildObjects(default_class, stream, loader)

  if len(definitions) < 1:
    raise yaml_errors.EmptyConfigurationFile()
  if len(definitions) > 1:
    raise yaml_errors.MultipleConfigurationFile()
  return definitions[0]



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#



"""Stub version of the urlfetch API, based on httplib."""

import gzip
import logging
import os
import socket
import ssl
import sys

from google.appengine.api import apiproxy_stub
from google.appengine.api import full_app_id
from google.appengine.api import urlfetch
from google.appengine.api import urlfetch_service_pb2



from google.appengine.runtime import apiproxy_errors
import six
from six.moves import http_client
from six.moves import range
from six.moves import urllib






MAX_REQUEST_SIZE = 10 << 20

MAX_RESPONSE_SIZE = 2 ** 25

MAX_REDIRECTS = urlfetch.MAX_REDIRECTS

REDIRECT_STATUSES = frozenset([
    http_client.MOVED_PERMANENTLY,
    http_client.FOUND,
    http_client.SEE_OTHER,
    http_client.TEMPORARY_REDIRECT,
])

PRESERVE_ON_REDIRECT = frozenset(['GET', 'HEAD'])

_API_CALL_DEADLINE = 900


_API_CALL_VALIDATE_CERTIFICATE_DEFAULT = True



_MAX_REQUEST_SIZE = 10485760







_UNTRUSTED_REQUEST_HEADERS = frozenset([
  'content-length',
  'host',
  'vary',
  'via',
  'x-forwarded-for',
])




_MAX_URL_LENGTH = 10240


def _SetupSSL(path):
  global CERT_PATH
  if os.path.exists(path):
    CERT_PATH = path
  else:
    CERT_PATH = None
    logging.warning('%s missing; without this urlfetch will not be able to '
                    'validate SSL certificates.', path)







def _IsAllowedPort(port):

  if port is None:
    return True
  try:
    port = int(port)
  except ValueError:
    return False




  if (port == 0 or
      (port >= 80 and port <= 90) or
      (port >= 440 and port <= 450) or
      port >= 1024):
    return True
  return False

def _IsLocalhost(host):
  """Determines whether 'host' points to localhost."""
  return host.startswith('localhost') or host.startswith('127.0.0.1')


def GetHeaders(msg, key):
  """Helper to get headers between python versions."""
  if six.PY2:
    return msg.getheaders(key)
  return msg.get_all(key)


class URLFetchServiceStub(apiproxy_stub.APIProxyStub):
  """Stub version of the urlfetch API to be used with apiproxy_stub_map."""

  THREADSAFE = True

  def __init__(self,
               service_name='urlfetch',
               urlmatchers_to_fetch_functions=None):
    """Initializer.

    Args:
      service_name: Service name expected for all calls.
      urlmatchers_to_fetch_functions: A list of two-element tuples.
        The first element is a urlmatcher predicate function that takes
        a url and determines a match. The second is a function that
        can retrieve result for that url. If no match is found, a url is
        handled by the default _RetrieveURL function.
        When more than one match is possible, the first match is used.
    """
    super(URLFetchServiceStub, self).__init__(service_name,
                                              max_request_size=MAX_REQUEST_SIZE)
    self._urlmatchers_to_fetch_functions = urlmatchers_to_fetch_functions or []
    self.http_proxy = None

  def _Dynamic_SetHttpProxy(self, request, response):
    self.http_proxy = (request.http_proxy_host, request.http_proxy_port)

  def _Dynamic_Fetch(self, request, response):
    """Trivial implementation of URLFetchService::Fetch().

    Args:
      request: the fetch to perform, a URLFetchRequest
      response: the fetch response, a URLFetchResponse
    """
    if len(request.Url) > _MAX_URL_LENGTH:
      logging.error('URL is too long: %s...', request.Url[:50])
      raise apiproxy_errors.ApplicationError(
          urlfetch_service_pb2.URLFetchServiceError.INVALID_URL)

    (protocol, host, _, _, _) = urllib.parse.urlsplit(request.Url)

    payload = None
    if request.Method == urlfetch_service_pb2.URLFetchRequest.GET:
      method = 'GET'
    elif request.Method == urlfetch_service_pb2.URLFetchRequest.POST:
      method = 'POST'
      payload = request.Payload
    elif request.Method == urlfetch_service_pb2.URLFetchRequest.HEAD:
      method = 'HEAD'
    elif request.Method == urlfetch_service_pb2.URLFetchRequest.PUT:
      method = 'PUT'
      payload = request.Payload
    elif request.Method == urlfetch_service_pb2.URLFetchRequest.DELETE:
      method = 'DELETE'
    elif request.Method == urlfetch_service_pb2.URLFetchRequest.PATCH:
      method = 'PATCH'
      payload = request.Payload
    else:
      logging.error('Invalid method: %s', request.Method)
      raise apiproxy_errors.ApplicationError(
          urlfetch_service_pb2.URLFetchServiceError.INVALID_URL)

    if payload is not None and len(payload) > _MAX_REQUEST_SIZE:
      raise apiproxy_errors.ApplicationError(
          urlfetch_service_pb2.URLFetchServiceError.PAYLOAD_TOO_LARGE)

    if not (protocol == 'http' or protocol == 'https'):
      logging.error('Invalid protocol: %s', protocol)
      raise apiproxy_errors.ApplicationError(
          urlfetch_service_pb2.URLFetchServiceError.INVALID_URL)

    if not host:
      logging.error('Missing host.')
      raise apiproxy_errors.ApplicationError(
          urlfetch_service_pb2.URLFetchServiceError.INVALID_URL)

    self._SanitizeHttpHeaders(_UNTRUSTED_REQUEST_HEADERS,
                              request.header)
    deadline = _API_CALL_DEADLINE
    if request.HasField('Deadline'):
      deadline = request.Deadline
    validate_certificate = _API_CALL_VALIDATE_CERTIFICATE_DEFAULT
    if request.HasField('MustValidateServerCertificate'):
      validate_certificate = request.MustValidateServerCertificate

    custom_fetch_function = self._GetCustomFetchFunction(request.Url)
    if custom_fetch_function:
      custom_fetch_function(request.Url, payload, method,
                            request.header, request, response,
                            follow_redirects=request.FollowRedirects,
                            deadline=deadline,
                            validate_certificate=validate_certificate)
    else:
      self._RetrieveURL(request.Url, payload, method,
                        request.header, request, response,
                        follow_redirects=request.FollowRedirects,
                        deadline=deadline,
                        validate_certificate=validate_certificate,
                        http_proxy=self.http_proxy)

  def _GetCustomFetchFunction(self, url):
    """Get the custom fetch function for a url.

    Args:
      url: A url to fetch from. str.

    Returns:
      A custom fetch function for this url, or None if no matching custom
      function is found.
    """
    for urlmatcher, fetch_function in self._urlmatchers_to_fetch_functions:
      if urlmatcher(url):
        return fetch_function
    return None

  @staticmethod
  def _RetrieveURL(url, payload, method, headers, request, response,
                   follow_redirects=True, deadline=_API_CALL_DEADLINE,
                   validate_certificate=_API_CALL_VALIDATE_CERTIFICATE_DEFAULT,
                   http_proxy=None):
    """Retrieves a URL over network.

    Args:
      url: String containing the URL to access.
      payload: Request payload to send, if any; None if no payload.
        If the payload is unicode, we assume it is utf-8.
      method: HTTP method to use (e.g., 'GET')
      headers: List of additional header objects to use for the request.
      request: A urlfetch_service_pb2.URLFetchRequest proto object from
          original request.
      response: A urlfetch_service_pb2.URLFetchResponse proto object to
          populate with the response data.
      follow_redirects: optional setting (defaulting to True) for whether or not
        we should transparently follow redirects (up to MAX_REDIRECTS)
      deadline: Number of seconds to wait for the urlfetch to finish.
      validate_certificate: If true, do not send request to server unless the
        certificate is valid, signed by a trusted CA and the hostname matches
        the certificate.
      http_proxy: Tuple of (hostname, port), where hostname is a string and port
        is an int, to use as the http proxy.

    Raises:
      Raises an apiproxy_errors.ApplicationError exception with
      INVALID_URL_ERROR in cases where:
        - The protocol of the redirected URL is bad or missing.
        - The port is not in the allowable range of ports.
      Raises an apiproxy_errors.ApplicationError exception with
      TOO_MANY_REDIRECTS in cases when MAX_REDIRECTS is exceeded
    """
    last_protocol = ''
    last_host = ''

    url = six.ensure_str(url, 'utf-8')

    if isinstance(payload, six.text_type):
      payload = six.ensure_str(payload, 'utf-8')

    for _ in range(MAX_REDIRECTS + 1):
      parsed = urllib.parse.urlsplit(url)
      protocol, host, path, query, _ = parsed

      port = parsed.port

      if not _IsAllowedPort(port):
        logging.error(
            'urlfetch received %s ; port %s is not allowed in production!',
            url, port)





        raise apiproxy_errors.ApplicationError(
            urlfetch_service_pb2.URLFetchServiceError.INVALID_URL)

      if protocol and not host:

        logging.error('Missing host on redirect; target url is %s', url)
        raise apiproxy_errors.ApplicationError(
            urlfetch_service_pb2.URLFetchServiceError.INVALID_URL)




      if not host and not protocol:
        host = last_host
        protocol = last_protocol


      if port == '0':
        host = host.replace(':0', '')









      adjusted_headers = {
          b'User-Agent': [(
              b'AppEngine-Google; (+http://code.google.com/appengine; appid: %s)'
              % full_app_id.get().encode())],
          b'Host': [six.ensure_binary(host)],
          b'Accept-Encoding': [b'gzip'],
      }
      if payload is not None:


        adjusted_headers[b'Content-Length'] = [
            six.ensure_binary(str(len(payload)))
        ]


      if method == 'POST' and payload:
        adjusted_headers[b'Content-Type'] = [
            b'application/x-www-form-urlencoded'
        ]

      passthrough_content_encoding = False
      for header in headers:





        header_key = six.ensure_binary(header.Key)
        header_value = six.ensure_binary(header.Value)
        if header_key.lower() == b'user-agent':
          adjusted_headers[header_key.title()] = [
              (b'%s %s' % (six.ensure_binary(header_value),
                           adjusted_headers[b'User-Agent'][0]))
          ]
        elif header_key.lower() == b'accept-encoding':
          passthrough_content_encoding = True
          adjusted_headers[header_key.title()] = [header_value]
        elif header_key.lower() == b'content-type':
          adjusted_headers[header_key.title()] = [header_value]
        else:
          adjusted_headers.setdefault(header_key, []).append(header_value)

      logging.debug(
          'Making HTTP request: host = %r, url = %r, payload = %.1000r, '
          'headers = %r', host, url, payload, adjusted_headers)
      try:
        proxy_host = None
        connection_kwargs = {'timeout': deadline}

        if protocol == 'http':
          connection_class = http_client.HTTPConnection
          default_port = 80

          if http_proxy and not _IsLocalhost(host):
            proxy_host = '%s:%d' % (http_proxy[0],
                                    http_proxy[1])
          elif os.environ.get('HTTP_PROXY') and not _IsLocalhost(host):
            _, proxy_host, _, _, _ = (
                urllib.parse.urlsplit(os.environ.get('HTTP_PROXY')))
        elif protocol == 'https':
          connection_class = http_client.HTTPSConnection
          if (validate_certificate and CERT_PATH):
            context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
            context.verify_mode = ssl.CERT_REQUIRED
            context.load_verify_locations(CERT_PATH)
            context.check_hostname = True
            connection_kwargs['context'] = context

          default_port = 443

          if os.environ.get('HTTPS_PROXY') and not _IsLocalhost(host):
            _, proxy_host, _, _, _ = (
                urllib.parse.urlsplit(os.environ.get('HTTPS_PROXY')))
        else:

          error_msg = 'Redirect specified invalid protocol: "%s"' % protocol
          logging.error(error_msg)
          raise apiproxy_errors.ApplicationError(
              urlfetch_service_pb2.URLFetchServiceError.INVALID_URL, error_msg)




        if (not validate_certificate and sys.version_info >= (2, 7, 9)
            and protocol == 'https'):
          connection_kwargs['context'] = ssl._create_unverified_context()

        if proxy_host:
          proxy_address, _, proxy_port = proxy_host.partition(':')
          connection = connection_class(
              proxy_address, proxy_port if proxy_port else default_port,
              **connection_kwargs)
          full_path = urllib.parse.urlunsplit((protocol, host, path, query, ''))

          if protocol == 'https':
            connection.set_tunnel(host)
        else:
          connection = connection_class(host, **connection_kwargs)
          full_path = urllib.parse.urlunsplit(('', '', path, query, ''))



        last_protocol = protocol
        last_host = host

        try:
          _SendRequest(connection, method, full_path, payload, adjusted_headers)
          http_response = connection.getresponse()
          if method == 'HEAD':
            http_response_data = ''
          else:
            http_response_data = http_response.read()
        finally:
          connection.close()
      except ssl.CertificateError as e:
        raise apiproxy_errors.ApplicationError(
            urlfetch_service_pb2.URLFetchServiceError.SSL_CERTIFICATE_ERROR,
            str(e))
      except ssl.SSLError as e:





        app_error = (
            urlfetch_service_pb2.URLFetchServiceError.DEADLINE_EXCEEDED
            if 'timed out' in str(e) else
            urlfetch_service_pb2.URLFetchServiceError.SSL_CERTIFICATE_ERROR)
        raise apiproxy_errors.ApplicationError(app_error, str(e))
      except socket.timeout as e:
        raise apiproxy_errors.ApplicationError(
            urlfetch_service_pb2.URLFetchServiceError.DEADLINE_EXCEEDED, str(e))
      except (http_client.error, socket.error, IOError) as e:
        raise apiproxy_errors.ApplicationError(
            urlfetch_service_pb2.URLFetchServiceError.FETCH_ERROR, str(e))

      if http_response.status >= 600:
        raise apiproxy_errors.ApplicationError(
            urlfetch_service_pb2.URLFetchServiceError.FETCH_ERROR,
            'Status %s unknown' % http_response.status)




      if http_response.status in REDIRECT_STATUSES and follow_redirects:

        url = http_response.getheader('Location', None)
        if url is None:
          error_msg = 'Missing "Location" header for redirect.'
          logging.error(error_msg)
          raise apiproxy_errors.ApplicationError(
              urlfetch_service_pb2.URLFetchServiceError.MALFORMED_REPLY,
              error_msg)



        if (http_response.status != http_client.TEMPORARY_REDIRECT and
            method not in PRESERVE_ON_REDIRECT):
          logging.warning('Received a %s to a %s. Redirecting with a GET',
                          http_response.status, method)
          method = 'GET'
          payload = None
      else:
        response.StatusCode = http_response.status
        if (http_response.getheader('content-encoding') == 'gzip' and
            not passthrough_content_encoding):
          gzip_stream = six.BytesIO(http_response_data)
          gzip_file = gzip.GzipFile(fileobj=gzip_stream)
          http_response_data = gzip_file.read()
        response.Content = six.ensure_binary(
            http_response_data[:MAX_RESPONSE_SIZE])




        key_set = set([key.lower() for key in http_response.msg.keys()])
        for header_key in key_set:
          header_values = GetHeaders(http_response.msg, header_key)
          if (header_key.lower() == 'content-encoding' and
              'gzip' in header_values and not passthrough_content_encoding):
            continue
          if header_key.lower() == 'content-length' and method != 'HEAD':
            header_values = [str(len(response.Content))]

          for header_value in header_values:
            response.header.add(Key=header_key, Value=header_value)

        if len(http_response_data) > MAX_RESPONSE_SIZE:
          response.ContentWasTruncated = True



        if request.Url != url:
          response.FinalUrl = url


        break
    else:
      error_msg = 'Too many repeated redirects'
      logging.error(error_msg)
      raise apiproxy_errors.ApplicationError(
          urlfetch_service_pb2.URLFetchServiceError.TOO_MANY_REDIRECTS,
          error_msg)

  def _SanitizeHttpHeaders(self, untrusted_headers, headers):
    """Cleans "unsafe" headers from the HTTP request, in place.

    Args:
      untrusted_headers: Set of untrusted headers names (all lowercase).
      headers: List of Header objects. The list is modified in place.
    """
    prohibited_headers = [h.Key for h in headers
                          if h.Key.lower() in untrusted_headers]
    if prohibited_headers:
      logging.warning('Stripped prohibited headers from URLFetch request: %s',
                      prohibited_headers)

      for index in reversed(range(len(headers))):
        if headers[index].Key.lower() in untrusted_headers:
          del headers[index]


def _SendRequest(connection, method, full_path, payload, headers):
  """Sends an HTTP request on a connection to the URL described by full_path.

  Compared to httplib.HTTPConnection's request method, this preserves all values
  for repeated headers.

  Args:
    connection: An instance or subclass of httplib.HTTPConnection.
    method: The string HTTP method name, eg 'GET'.
    full_path: The string full URL path for the request.
    payload: The string request payload to send.
    headers: A dict of headers to send with the request. The dict maps string
      header names to lists of associated header values.
  """
  connection.connect()
  header_names = [name.lower() for name in headers]
  connection.putrequest(
      method,
      full_path,
      skip_host=b'host' in header_names,
      skip_accept_encoding=b'accept-encoding' in header_names)
  for header, values in six.iteritems(headers):
    for value in values:
      connection.putheader(header, value)
  if payload is not None and b'content-length' not in header_names:
    connection._set_content_length(payload)
  connection.endheaders(payload)
